#!/bin/python3

import argparse
import subprocess


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--no-cache",
        help="Disable using cache for building the docker image",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "PYTHON_VERSION",
        help="The python version to build the image with.",
        choices=["3.10", "3.11", "3.12"],
    )
    parser.add_argument(
        "FINESSE_REF",
        help="Optional argument specifying the finesse git reference. Defaults to 'develop'",
        nargs="?",
        default="develop",
    )
    args = parser.parse_args()
    print(
        f"Building finesse cython debugging image for Python {args.PYTHON_VERSION} and finesse reference {args.FINESSE_REF}"
    )
    cmd1 = f"docker build -t gdb-with-python:{args.PYTHON_VERSION} --build-arg PYTHON_VERSION={args.PYTHON_VERSION} -f gdb_python.Dockerfile ."
    print(cmd1)
    subprocess.run(cmd1, shell=True, check=True)
    cmd2 = f"docker build {'--no-cache' if args.no_cache else ''} -t finesse-cython-debug:{args.PYTHON_VERSION} --build-arg PYTHON_VERSION={args.PYTHON_VERSION} --build-arg FINESSE_REF={args.FINESSE_REF} ."
    print(cmd2)
    subprocess.run(cmd2, shell=True, check=True)


if __name__ == "__main__":
    main()



"""Transmissive optical components which focus or disperse light beams."""

import logging
import numpy as np

from finesse.components.general import Connector, InteractionType, NoiseType
from finesse.components.node import NodeDirection, NodeType
from finesse.parameter import float_parameter
from finesse.utilities import refractive_index
from finesse.tracing import abcd
from finesse.symbols import Matrix, Constant
from finesse.exceptions import FinesseException

from finesse.env import warn
from finesse.warnings import UnreasonableComponentValueWarning

from abc import ABC

LOGGER = logging.getLogger(__name__)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class BaseLens(Connector, ABC):
    """Represents a thin lens optical component with an associated focal length.

    Notes
    -----
    The specified focal length `f` is only accurate when the lens is attached to spaces with index
    of refraction close to 1. This component exists so that one can use the intuitive focal length
    parameter instead of having to set radii of curvature as with e.g. :class:`.Mirror`.

    Parameters
    ----------
    name : str
        Name of newly created lens.

    f : float, optional
        Focal length of the lens in metres; defaults to infinity.

    Attributes
    ----------
    OPD_map : :class:`finesse.knm.maps.Map`
        A map that is used to describe the transverse spatial amplitude and phase
        variations beyond a simple lensing. Typically the map applied is describing
        the dnr/dT temperature effects in some substrate.
    """

    def __init__(
        self,
        name,
    ):
        super().__init__(name)
        self.OPD_map = None

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

    def optical_equations(self):
        if self._model._settings.is_modal:
            return {
                f"{self.name}.P1i_P2o": Matrix(f"{self.name}.K12"),
                f"{self.name}.P2i_P1o": Matrix(f"{self.name}.K21"),
            }
        else:
            return {
                f"{self.name}.P1i_P2o": Constant(1),
                f"{self.name}.P2i_P1o": Constant(1),
            }

    def _check_f(self, value):
        if value == 0:
            raise ValueError("Focal length of lens must be non-zero.")

        # check for unreasonable focal length and suggest alternative
        if abs(value) < 10e-3:
            warn(
                f"Lens '{self.name}' has a small focal length. The 'lens' ABCD "
                "matrix is valid for focal lengths >> thickness of the lens."
                "This lens may be better modelled as two mirrors and a space.",
                UnreasonableComponentValueWarning,
            )

        return value

    def _resymbolise_ABCDs(self):
        self._symbolise_ABCDs("x")
        self._symbolise_ABCDs("y")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
        allow_reverse=False,
    ):
        r"""Returns the ABCD matrix of the lens for the specified coupling.

        .. _fig_abcd_lens_transmission:
        .. image:: /images/lens/abcd_lenst.*
            :align: center

        This is given by,

        .. math::
            M = \begin{pmatrix}
                    1 & 0 \\
                    -\frac{1}{f} & 1
                \end{pmatrix},

        where :math:`f` is the focal length of the lens.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.
        """
        return super().ABCD(
            from_node, to_node, direction, symbolic, copy, retboth, allow_reverse
        )

    def _fill_optical_matrix(self, ws, matrix, connections):
        for freq in matrix.optical_frequencies.frequencies:
            with matrix.component_edge_fill3(
                ws.owner_id,
                connections.P1i_P2o_idx,
                freq.index,
                freq.index,
            ) as mat:
                mat[:] = ws.K12.data

            with matrix.component_edge_fill3(
                ws.owner_id,
                connections.P2i_P1o_idx,
                freq.index,
                freq.index,
            ) as mat:
                mat[:] = ws.K21.data

    def _fill_carrier(self, ws):
        self._fill_optical_matrix(ws, ws.sim.carrier, ws.carrier.connections)

    def _fill_signal(self, ws):
        self._fill_optical_matrix(ws, ws.sim.signal, ws.signal.connections)

    def _get_baselens_workspace(self, sim, ws_type):
        from finesse.components.modal.lens import lens_fill_qnoise

        _, is_changing = self._eval_parameters()
        refill = sim.is_component_in_mismatch_couplings(self) or len(is_changing)

        ws = ws_type(self, sim)
        # This assumes that nr1/nr2 cannot change during a simulation
        ws.nr1 = refractive_index(self.p1)
        ws.nr2 = refractive_index(self.p2)
        # TODO ddb refractive index should be equal on
        # both sides of the lens as we are using the thin
        # lens approximation
        if not np.allclose(ws.nr1, ws.nr2, rtol=1e-13):
            raise FinesseException(
                "Refractive index on both sides of the lens must be equal"
            )

        ws.carrier.add_fill_function(self._fill_carrier, refill)
        ws.signal.add_fill_function(self._fill_signal, refill)

        if sim.is_modal:
            ws.abcd_x = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)
            ws.abcd_y = self.ABCD(self.p1.i, self.p2.o, "y", copy=False)

            # Set the coupling matrix information
            # ABCDs are same in each direction
            ws.set_knm_info(
                "P1i_P2o",
                abcd_x=ws.abcd_x,
                abcd_y=ws.abcd_y,
                nr_from=ws.nr1,
                nr_to=ws.nr2,
                is_transmission=True,
                apply_map=self.OPD_map,
                map_phase_factor=1,
                map_fliplr=False,
            )
            # TODO nr reversed here for now until it's forced to be same on both sides
            ws.set_knm_info(
                "P2i_P1o",
                abcd_x=ws.abcd_x,
                abcd_y=ws.abcd_y,
                nr_from=ws.nr2,
                nr_to=ws.nr1,
                is_transmission=True,
                apply_map=self.OPD_map,
                map_phase_factor=1,
                map_fliplr=True,
            )

            if sim.signal:
                ws.signal.set_fill_noise_function(NoiseType.QUANTUM, lens_fill_qnoise)
        return ws


@float_parameter(
    "f",
    "Focal length",
    validate="_check_f",
    units="m",
    is_geometric=True,
)
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Lens(BaseLens):
    """Represents a thin lens optical component with an associated focal length.

    Notes
    -----
    The specified focal length `f` is only accurate when the lens is attached to spaces with index
    of refraction close to 1. This component exists so that one can use the intuitive focal length
    parameter instead of having to set radii of curvature as with e.g. :class:`.Mirror`.

    Parameters
    ----------
    name : str
        Name of newly created lens.

    f : float, optional
        Focal length of the lens in metres; defaults to infinity.

    Attributes
    ----------
    OPD_map : :class:`finesse.knm.maps.Map`
        A map that is used to describe the transverse spatial amplitude and phase
        variations beyond a simple lensing. Typically the map applied is describing
        the dnr/dT temperature effects in some substrate.
    """

    def __init__(self, name, f=np.inf):
        super().__init__(name)
        self.f = f
        self.OPD_map = None

    def _symbolise_ABCDs(self, direction):
        M_sym = abcd.lens(self.f.ref)

        # Matrices same for both node couplings
        self.register_abcd_matrix(
            M_sym,
            (self.p1.i, self.p2.o, direction),
            (self.p2.i, self.p1.o, direction),
        )

    def _get_workspace(self, sim):
        from finesse.simulations.sparse.simulation import SparseMatrixSimulation

        if isinstance(sim, SparseMatrixSimulation):
            from finesse.components.modal.lens import LensWorkspace

            return self._get_baselens_workspace(sim, LensWorkspace)


@float_parameter(
    "fx",
    "Focal length (x-z plane)",
    validate="_check_f",
    units="m",
    is_geometric=True,
)
@float_parameter(
    "fy",
    "Focal length (y-z plane)",
    validate="_check_f",
    units="m",
    is_geometric=True,
)
# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class AstigmaticLens(BaseLens):
    """Represents a thin astigmatic lens optical component with an associated focal
    lengths.

    Notes
    -----
    The specified focal length `f` is only accurate when the lens is attached to spaces with index
    of refraction close to 1. This component exists so that one can use the intuitive focal length
    parameter instead of having to set radii of curvature as with e.g. :class:`.Mirror`.

    Parameters
    ----------
    name : str
        Name of newly created lens.

    fx : float, optional
        Focal length in x-z plane of the lens in metres; defaults to infinity.
    fy : float, optional
        Focal length of y-z plane the lens in metres; defaults to infinity.

    Attributes
    ----------
    OPD_map : :class:`finesse.knm.maps.Map`
        A map that is used to describe the transverse spatial amplitude and phase
        variations beyond a simple lensing. Typically the map applied is describing
        the dnr/dT temperature effects in some substrate.
    """

    def __init__(self, name, fx=np.inf, fy=np.inf):
        super().__init__(name)
        self.fx = fx
        self.fy = fy
        self.OPD_map = None

    def _symbolise_ABCDs(self, direction):
        if direction == "x":
            M_sym = abcd.lens(self.fx.ref)
        elif direction == "y":
            M_sym = abcd.lens(self.fy.ref)
        else:
            raise ValueError(f"Invalid direction: {direction}")

        # Matrices same for both node couplings
        self.register_abcd_matrix(
            M_sym,
            (self.p1.i, self.p2.o, direction),
            (self.p2.i, self.p1.o, direction),
        )

    def _get_workspace(self, sim):
        from finesse.simulations.sparse.simulation import SparseMatrixSimulation

        if isinstance(sim, SparseMatrixSimulation):
            from finesse.components.modal.lens import AstigmaticLensWorkspace

            return self._get_baselens_workspace(sim, AstigmaticLensWorkspace)



from finesse.cymath.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.cymath cimport complex_t
from finesse.simulations.base cimport ModelSettings
from finesse.frequency cimport frequency_info_t
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.simulations.workspace cimport GouyFuncWrapper
from finesse.element_workspace cimport BaseCValues

from finesse.cyexpr cimport (
    cy_expr,
    cy_expr_new,
    cy_expr_init,
    cy_expr_free,
    cy_expr_eval,
)

from cpython.ref cimport PyObject

import numpy as np
cimport numpy as np


cdef struct space_optical_connections:
    PyObject** P1i_P2o
    PyObject** P2i_P1o


cdef struct space_signal_connections:
    PyObject*** SIGAMP_P1o
    PyObject*** SIGAMP_P2o
    PyObject*** SIGPHS_P1o
    PyObject*** SIGPHS_P2o
    PyObject*** H_P1o
    PyObject*** H_P2o


cdef class SpaceOpticalConnections:
    cdef public:
        int P1i_P2o_idx
        int P2i_P1o_idx
    cdef readonly:
        SubCCSView1DArray P1i_P2o
        SubCCSView1DArray P2i_P1o
    cdef:
        space_optical_connections opt_ptrs


cdef class SpaceSignalConnections(SpaceOpticalConnections):
    cdef public:
        int SIGPHS_P1o_idx
        int SIGPHS_P2o_idx
        int SIGAMP_P1o_idx
        int SIGAMP_P2o_idx
        int H_P1o_idx
        int H_P2o_idx
    cdef readonly:
        SubCCSView2DArray SIGPHS_P1o
        SubCCSView2DArray SIGPHS_P2o
        SubCCSView2DArray SIGAMP_P1o
        SubCCSView2DArray SIGAMP_P2o
        SubCCSView2DArray H_P1o
        SubCCSView2DArray H_P2o
    cdef:
        space_signal_connections sig_ptrs


cdef class SpaceValues(BaseCValues):
    cdef public:
        double L
        double nr

        # Holds user and bt gouy values
        double user_gouy_x
        double user_gouy_y
        double computed_gouy_x
        double computed_gouy_y


cdef class SpaceWorkspace(ConnectorWorkspace):
    cdef public:
        # Indices (in sim._c_node_info and sim.trace) of nodes
        Py_ssize_t P1i_id
        Py_ssize_t P1o_id
        Py_ssize_t P2i_id
        Py_ssize_t P2o_id

        Py_ssize_t car_p1o_idx
        Py_ssize_t car_p2o_idx

        # ABCD matrix view (note same for all couplings)
        double[:, ::1] abcd

        # Flags to use user or bt gouy phase
        bint use_user_gouy_x
        bint use_user_gouy_y
    cdef:
        SpaceOpticalConnections sco
        SpaceSignalConnections scs
        SpaceValues sv

        # Only element B of ABCD matrix can change, store pointer
        # to compiled changing expression for this element here
        cy_expr* sym_abcd_B
        complex_t* couplings

        Py_ssize_t car_p1o_rhs_idx
        Py_ssize_t car_p2o_rhs_idx
        Py_ssize_t car_p_num_hom
        bint strain_signal_enabled
        bint phase_signal_enabled
        bint amp_signal_enabled

    cpdef update_parameter_values(self)


cdef object c_space_carrier_fill(ConnectorWorkspace cws)
cdef void space_fill_optical_2_optical(
        space_optical_connections *conn,
        SpaceWorkspace ws,
        frequency_info_t *freq,
        double pre_factor,
        double gouy_x,
        double gouy_y,
        bint zero_tem00_gouy,
        int[:, ::1] homs_view
    ) noexcept



"""Finesse Sphinx extension.

This provides the `kat:command`, `kat:element` and `kat:analysis` directives and indices
to provide a means to automatically document and cross-reference KatScript instructions.

Loosely based on https://www.sphinx-doc.org/en/master/development/tutorials/recipe.html
and https://github.com/click-contrib/sphinx-click/blob/master/sphinx_click/ext.py.

Short description of how Sphinx, Docutils and this module works:

Sphinx ultimately relies upon the Python library `docutils` to generate the Finesse
documentation. This provides a way to define the document structure independently from
the target final format. By defining the document in this way, Sphinx (via docutils) can
generate the various output formats (HTML, PDF, etc.) as required. The docutils library
provides most of the low-level elements for building the document such as those that
represent paragraphs, section headers, lists, etc. Sphinx adds some higher order classes
that represent more complex structures such as class definitions, as well as providing
support for generating indices, contents lists, cross-references, etc.

Docutils documents are a tree of "nodes", similar to an HTML document. The node types
that represent various types of textual, structural and decorative elements are defined
in :mod:`docutils.nodes` and :mod:`sphinx.addnodes`.

The `finesse_sphinx` package registers several new Sphinx objects for referencing and
generating documentation for items such as KatScript elements. These are special classes
that interact with Sphinx while the Finesse documentation is being generated to inject
additional content and references and to create additional index entries.

Documentation for building custom Sphinx extensions such as those in this module is
unfortunately very limited. The main reference is the Sphinx source code itself,
particularly the `autodoc` extension, which implements similar behaviours to those of
the extensions here.

Author: Sean Leavey
"""

import abc
from collections import defaultdict
from inspect import getabsfile

from docutils import nodes
from docutils.statemachine import ViewList
from finesse.script.spec import KATSPEC
from sphinx import addnodes
from sphinx.domains import Domain, Index
from sphinx.roles import XRefRole
from sphinx.util.docutils import ReferenceRole, SphinxDirective
from sphinx.util.nodes import make_refnode, nested_parse_with_titles

__version__ = "0.9.1"


def kat_syntax(adapter):
    """Build kat syntax string for `adapter`."""
    return adapter.documenter.syntax(KATSPEC, adapter)


class IssueRole(ReferenceRole):
    """Support for referencing issues in the Finesse source repository.

    Based on Sphinx's PEP role.
    """

    def run(self):
        # Add an index entry.
        target_id = "index-%s" % self.env.new_serialno("index")
        entries = [("single", f"Finesse issues; #{self.target}", target_id, "", None)]

        index = addnodes.index(entries=entries)
        target = nodes.target("", "", ids=[target_id])
        self.inliner.document.note_explicit_target(target)

        try:
            refuri = self.build_uri()
            reference = nodes.reference(
                "", "", internal=False, refuri=refuri, classes=["finesse-issue"]
            )
            if self.has_explicit_title:
                reference += nodes.strong(self.title, self.title)
            else:
                title = f"#{self.title}"
                reference += nodes.strong(title, title)
        except ValueError:
            msg = self.inliner.reporter.error(
                f"invalid Finesse issue {self.target}", line=self.lineno
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)
            return [prb], [msg]

        return [index, target, reference], []

    def build_uri(self):
        return self.config.finesse_issue_uri % int(self.target)


class SourceRole(ReferenceRole):
    """Support for referencing source code in the Finesse source repository.

    Based on Sphinx's PEP role.
    """

    def run(self):
        # Add an index entry.
        target_id = "index-%s" % self.env.new_serialno("index")
        entries = [
            ("single", f"Finesse source code; {self.target}", target_id, "", None)
        ]

        index = addnodes.index(entries=entries)
        target = nodes.target("", "", ids=[target_id])
        self.inliner.document.note_explicit_target(target)

        try:
            refuri = self.build_uri()
            reference = nodes.reference(
                "", "", internal=False, refuri=refuri, classes=["finesse-source-code"]
            )
            reference += nodes.strong(self.title, self.title)
        except ValueError:
            msg = self.inliner.reporter.error(
                f"invalid Finesse source code path {self.target}", line=self.lineno
            )
            prb = self.inliner.problematic(self.rawtext, self.rawtext, msg)
            return [prb], [msg]

        return [index, target, reference], []

    def build_uri(self):
        if self.target.startswith("/"):
            # This is relative to the project root.
            return self.config.finesse_source_root_uri % self.target.lstrip("/")
        # This is relative to the Finesse package root.
        return self.config.finesse_source_package_uri % self.target


class InstructionDirective(SphinxDirective, metaclass=abc.ABCMeta):
    """A custom Sphinx directive that describes a KatScript instruction."""

    description = "Instruction"
    has_content = True
    required_arguments = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        directive = self.arguments[0]
        try:
            self.adapter = KATSPEC.directives[directive]
        except KeyError:
            raise ValueError(
                f"directive {directive} doesn't exist in KatScript language spec"
            )

    def _description_slug(self):
        """Description type, usable as HTML attribute."""
        return nodes.fully_normalize_name(self.description)

    def _alias_slug(self, alias):
        return nodes.make_id(f"kat-{self._description_slug()}-{alias}")

    def _params(self):
        required_params = []
        optional_params = []

        descriptions = self.adapter.documenter.argument_descriptions()
        for param_name, (param_type, param_doc) in descriptions.items():
            param = param_name, param_doc
            if "optional" in param_type:
                optional_params.append(param)
            else:
                required_params.append(param)

        return required_params, optional_params

    def run(self):
        """Generate the nodes that represent this directive.

        This generates the signature line (directive aliases), summary, parameter lists
        and any other content specified in the restructuredText document.
        """
        doc_type = self.adapter.documenter.doc_type

        try:
            source = getabsfile(doc_type)
        except TypeError:
            # This may be a Cythonised target. Use the string name instead.
            source = doc_type.__name__

        kat = self.env.get_domain("kat")

        if ":" not in self.name:
            raise ValueError(
                f"incorrect name format '{self.name}'; must be of the form 'kat:thing'"
            )

        self.domain, self.objtype = self.name.split(":", 1)

        # Create the index target for the instruction, just before the signature in the
        # document.
        index_node = addnodes.index(entries=[])

        node = addnodes.desc()
        node.document = self.state.document
        node["domain"] = self.domain
        node["objtype"] = self.objtype
        node["noindex"] = noindex = "noindex" in self.options
        node["classes"].append(self.domain)
        node["classes"].append(node["objtype"])

        # Create the signature and index entries for each alias.
        signature_node = addnodes.desc_signature("", "", is_multiline=True)
        self.set_source_info(signature_node)
        node += signature_node
        for alias in self.adapter.aliases:
            is_primary = alias == self.adapter.full_name

            # All signatures have anchors due to the signature_node["ids"] line below,
            # but add_permalink shows a "#" next to the signature that users can copy.
            signature_line = addnodes.desc_signature_line("", add_permalink=is_primary)
            signature_line += addnodes.desc_name(text=alias)
            signature_node += signature_line

            # Map index entry.
            if not noindex:
                anchor = self._alias_slug(alias)

                # Register an anchor for this alias.
                signature_node["ids"].append(anchor)

                kat.add_instruction(anchor, self, alias)

                if "noindexentry" not in self.options:
                    # Add an index entry for this alias.
                    index_text = f"{alias} (KatScript {self.description})"
                    entry = ("single", index_text, anchor, "", None)
                    index_node["entries"].append(entry)

        # Create a node to contain this instruction's documentation.
        content_node = addnodes.desc_content()
        self.set_source_info(content_node)
        node += content_node

        ##
        # The summary.
        ##

        summary = self.adapter.documenter.summary()
        if summary is None:
            summary = ""
        # Parse the docstring summary.
        summary_rst = ViewList()
        for line in summary.splitlines():
            summary_rst.append(line, source)
        self.state.nested_parse(summary_rst, 0, content_node)

        extended_summary = self.adapter.documenter.extended_summary()
        if extended_summary is None:
            extended_summary = ""
        # Parse the docstring extended summary.
        extended_summary_rst = ViewList()
        for line in extended_summary.splitlines():
            extended_summary_rst.append(line, source)
        self.state.nested_parse(extended_summary_rst, 0, content_node)

        # Field list (mapping of entries like "Syntax" to corresponding content).
        fields = nodes.field_list()
        content_node += fields

        ##
        # The syntax.
        ##

        syntax = kat_syntax(self.adapter)
        syntax_code_block = nodes.literal_block("", syntax)
        # Doesn't work with variadic args.
        if "*" not in syntax:
            syntax_code_block["language"] = "katscript"
        syntax_code_block["classes"].append("kat-instruction-syntax")
        syntax_node = nodes.field()
        syntax_node += nodes.field_name("", "Syntax")
        syntax_node += nodes.field_body("", syntax_code_block)
        fields += syntax_node

        ##
        # The parameters.
        ##

        required_params, optional_params = self._params()

        def param_field(desc, params):
            params_node = nodes.field()
            params_node += nodes.field_name("", desc)
            field_body = nodes.field_body()
            params_node += field_body

            for pnames, description in params:
                # Make the parameter(s) appear as literal text.
                param_names = ""
                for i, pname in enumerate(pnames.split(",")):
                    if i > 0:
                        param_names += ", "
                    param_names += f"``{pname.strip()}``"

                description = f"{param_names}: {description}"

                # Parse the description rsT.
                param_rst = ViewList()
                for line in description.strip().splitlines():
                    param_rst.append(line, source)
                self.state.nested_parse(param_rst, 0, field_body)

            return params_node

        if required_params:
            fields += param_field("Required", required_params)
        if optional_params:
            fields += param_field("Optional", optional_params)

        ##
        # Any extra rST specified in the rST file.
        ##

        content = ViewList()
        for line in self.content:
            content.append(line, source)
        content.append("", source)
        nested_parse_with_titles(self.state, content, content_node)

        return [index_node, node]


class CommandDirective(InstructionDirective):
    """A custom directive that describes a kat script command."""

    description = "Command"


class ElementDirective(InstructionDirective):
    """A custom directive that describes a kat script element."""

    description = "Element"


class AnalysisDirective(InstructionDirective):
    """A custom directive that describes a kat script analysis."""

    description = "Analysis"


class KatIndex(Index, metaclass=abc.ABCMeta):
    """A custom index that creates an instruction matrix."""

    index_type = None

    def generate(self, docnames=None):
        content = defaultdict(list)

        # Sort the list of instructions in alphabetical order.
        instructions = self.domain.get_objects()
        instructions = sorted(instructions, key=lambda instruction: instruction[0])

        # Generate the expected output, shown below, from the above using the first
        # letter of the recipe as a key to group thing.
        #
        # name, subtype, docname, anchor, extra, qualifier, description
        for _, dispname, typ, docname, anchor, _ in instructions:
            if typ != self.index_type:
                continue

            display_name = self.format_display_name(dispname)
            description = typ

            # The key is the index (in this case, the first character).
            content[display_name[0].lower()].append(
                (display_name, 0, docname, anchor, docname, "", description)
            )

        # Convert the dict to the sorted list of tuples expected.
        content = sorted(content.items())

        return content, True

    def format_display_name(self, name):
        return name


class CommandIndex(KatIndex):
    """A custom index that creates a command matrix."""

    name = "commandindex"
    localname = "Command Index"
    shortname = "Command"
    index_type = "Command"


class ElementIndex(KatIndex):
    """A custom index that creates an element matrix."""

    name = "elementindex"
    localname = "Element Index"
    shortname = "Element"
    index_type = "Element"


class AnalysisIndex(KatIndex):
    """A custom index that creates an analysis matrix."""

    name = "analysisindex"
    localname = "Analysis Index"
    shortname = "Analysis"
    index_type = "Analysis"


class KatDomain(Domain):
    name = "kat"
    label = "Kat Domain"
    roles = {
        "command": XRefRole(),
        "element": XRefRole(),
        "analysis": XRefRole(),
    }
    directives = {
        "command": CommandDirective,
        "element": ElementDirective,
        "analysis": AnalysisDirective,
    }
    indices = {
        CommandIndex,
        ElementIndex,
        AnalysisIndex,
    }
    initial_data = {
        "instructions": [],  # Object list.
    }

    def get_full_qualified_name(self, node):
        return "{}.{}".format("instruction", node.arguments[0])

    def get_objects(self):
        yield from self.data["instructions"]

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        # Match the specified reference to an alias of the available kat script
        # directives.
        if typ == "command":
            match = [
                (docname, anchor)
                for name, sig, otyp, docname, anchor, prio in self.get_objects()
                if (otyp == "Command" and target in KATSPEC.commands[sig].aliases)
            ]
        elif typ == "element":
            match = [
                (docname, anchor)
                for name, sig, otyp, docname, anchor, prio in self.get_objects()
                if (otyp == "Element" and target in KATSPEC.elements[sig].aliases)
            ]
        elif typ == "analysis":
            match = [
                (docname, anchor)
                for name, sig, otyp, docname, anchor, prio in self.get_objects()
                if (otyp == "Analysis" and target in KATSPEC.analyses[sig].aliases)
            ]
        else:
            match = []

        if len(match) > 0:
            todocname = match[0][0]
            targ = match[0][1]

            return make_refnode(builder, fromdocname, todocname, targ, contnode, targ)
        else:
            print(f"Did not find kat script instruction for xref '{target}'")
            return None

    def add_instruction(self, anchor, instruction, signature):
        """Add a new instruction to the domain."""
        name = "{}.{}".format("instruction", signature)

        # name, dispname, type, docname, anchor, priority
        self.data["instructions"].append(
            (name, signature, instruction.description, self.env.docname, anchor, 0)
        )

    def merge_domaindata(self, docnames, otherdata):
        for instruction in otherdata["instructions"]:
            if instruction not in self.data["instructions"]:
                self.data["instructions"].append(instruction)


def setup(app):
    app.add_config_value(
        "finesse_issue_uri",
        "https://gitlab.com/ifosim/finesse/finesse3/-/issues/%d",
        "html",
    )
    app.add_config_value(
        "finesse_source_root_uri",
        "https://gitlab.com/ifosim/finesse/finesse3/-/tree/master/%s",
        "html",
    )
    app.add_config_value(
        "finesse_source_package_uri",
        "https://gitlab.com/ifosim/finesse/finesse3/-/tree/master/src/finesse/%s",
        "html",
    )
    app.add_role("issue", IssueRole())
    app.add_role("source", SourceRole())
    app.add_domain(KatDomain)

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""A collection of methods to compute overlap integrals for modal scattering matrices.
Essentially this involves solving the following integral

.. math::
    K_{abnm} = \iint^{\infty}_{-\infty} U_{nm}(x,y,q_i) M(x,y) U^*_{ab}(x,y,q_o) \, dy \, dx

:math:`U_nm` is the initial modes in the basis :math:`q_i` we are converting from and
:math:`U_ab` are the target modes in a basis :math:`q_o` we are projecting into.

TODO
----
Should explore if decomposing compute_map_knm_matrix_riemann_optimised into real and imaginary
integrals might be faster. In cases where q_in == q_out then integrals are real, apart from
the map component which can be complex.

Explore use of zgemm3m which is 25% faster than zgemm

Probably look into using CUDA if necessary for more speed.
"""
import cython
import numpy as np
cimport numpy as np
from finesse.cymath cimport complex_t
from scipy.linalg.cython_blas cimport zgemm
from scipy.integrate import newton_cotes
from ..cymath.homs cimport unm_workspace, unm_factor_store, u_n__fast, u_m__fast


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef void outer_conj_product(complex_t[:,::1] U, complex_t[:,:,::1] result) except *:
    """Computes U * U^C and returns the output into the result array.
    Result array must be of shape (N,N,M) where U is shape (N,M).
    """
    cdef:
        Py_ssize_t i, j, k
        double a, b, c, d

    A = U.shape[0]
    B = U.shape[1]
    if not (result.shape[0] == result.shape[1] == A) or not (result.shape[2] == B):
        raise RuntimeError(f"result array wrong shape should be ({A,A,B})")

    for i in range(A):
        for j in range(A):
            for k in range(B):
                a = U[i, k].real
                b = U[i, k].imag
                c = U[j, k].real
                d = U[j, k].imag
                result[i, j, k] = a*c + b*d + 1j*(b*c - a*d) # U[i, k] * conj(U[j, k])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void c_outer_conj_product(
        Py_ssize_t Nm,
        Py_ssize_t Ns,
        complex_t *U, # complex[Nm, Ns] C-ordered
        complex_t *result  # complex[Nm, Nm, Ns] C-ordered
    ) noexcept nogil:
    """Computes U * U^C and returns the output into the result array.
    Result array must be of shape (Nm,Nm,Ns) where U is shape (Nm,Ns).
    """
    cdef:
        Py_ssize_t i, j, k
        double a, b, c, d

    for i in range(Nm):
        for j in range(Nm):
            for k in range(Ns):
                a = U[i*Ns + k].real #U[i, k].real
                b = U[i*Ns + k].imag
                c = U[j*Ns + k].real
                d = U[j*Ns + k].imag
                result[i*Nm*Ns + j*Ns + k] = a*c + b*d + 1j*(b*c - a*d) # U[i, k] * conj(U[j, k])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cpdef void outer_conj_product_2(complex_t[:,::1] U1, complex_t[:,::1] U2, complex_t[:,:,::1] result) except *:
    """Computes U1 * U2**C and returns the output into the result array.
    Result array must be of shape (N,N,M) where U is shape (N,M).
    """
    cdef:
        Py_ssize_t i, j, k
        double a, b, c, d

    A = U1.shape[0]
    B = U1.shape[1]
    if not (result.shape[0] == result.shape[1] == A) or not (result.shape[2] == B):
        raise RuntimeError(f"result array wrong shape should be ({A,A,B})")

    for i in range(A):
        for j in range(A):
            for k in range(B):
                a = U1[i, k].real
                b = U1[i, k].imag
                c = U2[j, k].real
                d = U2[j, k].imag
                result[i, j, k] = a*c + b*d + 1j*(b*c - a*d) # U1[i, k] * conj(U2[j, k])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void c_outer_conj_product_2(
        Py_ssize_t Nm,
        Py_ssize_t Ns,
        complex_t* U1,
        complex_t* U2,
        complex_t* result
    ) noexcept nogil:
    """Computes U1 * U2**C and returns the output into the result array.
    Result array must be of shape (N,N,M) where U is shape (N,M).
    """
    cdef:
        Py_ssize_t i, j, k
        double a, b, c, d

    for i in range(Nm):
        for j in range(Nm):
            for k in range(Ns):
                a = U1[i*Ns + k].real
                b = U1[i*Ns + k].imag
                c = U2[j*Ns + k].real
                d = U2[j*Ns + k].imag
                result[i*Ns*Nm + j*Ns + k] = a*c + b*d + 1j*(b*c - a*d) # U1[i, k] * conj(U2[j, k])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void update_U_xy_array(
        double *x, Py_ssize_t Nx, double *y, Py_ssize_t Ny,
        complex_t *Un, complex_t *Um, Py_ssize_t Nm,
        unm_workspace *ws, unm_factor_store *store
    ) noexcept nogil:
    """Method to fill Un and Um arrays up to some mode index Nm
    Parameters
    ----------
    x : *double
        Pointer to start of contiguous x array
    Nx : int
        Size of x array
    y : *double
        Pointer to start of contiguous y array
    Ny : int
        Size of y array
    Nm : int
        Max number of modes to calculate up to
    ws
        Pointer to allocated and initialised unm_workspace
    store
        Pointer to allocated and initialised unm_factor_store
    """
    cdef Py_ssize_t i, j
    for i in range(Nm):
        for j in range(Nx):
            Un[i*Nx + j] = u_n__fast(ws, store, i, x[j])

    for i in range(Nm):
        for j in range(Ny):
            Um[i*Ny + j] = u_m__fast(ws, store, i, y[j])


cpdef compute_map_scattering_coeffs_riemann_optimised(
            double dA,
            complex_t[:,::1] Z,
            complex_t[:,:,::1] Unn_,
            complex_t[:,:,::1] Umm_,
            complex_t[:,:,::1] tmp,
            complex_t[:,:,:,::1] result,
        ) :
    """Calculates a mode scattering matrix using a Riemann sum. This method
    uses an computationally optimised approach making use of fast BLAS
    functions. This requires the input modes to be specified in specific
    formats and memory layouts. What this functions computes is the following
    via a Riemann
    sum:

    .. math::
        K_{abnm} = \int^{\infty}_{-\infty}  u_{m}(y,q^y_i) u^*_{b}(y,q^y_o) \Bigg[ \int^{\infty}_{-\infty} Z(x,y) u_{n}(x,q^x_i) u^*_{a}(x,q^x_o) \, dx \Bigg] \, dy

    This integral is not actually performed to infinity, it is bound by the
    dimensions of the discretised map :math:`Z`. The map bound and uniform
    discretisation must be chosen to efficiently sample the size of any of
    the beams and maximum mode order being used.

    Due to the optimised calculation method, the result indexing is not
    [a,b,n,m] but [m,a,n,b]. To convert it back to a more usable
    indexing use:

    >>> result = np.transpose(result, (2,1,0,3))

    Notes
    -----
    Nx - number of x samples
    Ny - number of y samples
    Nm - number of modes (n, m) being calculated

    Parameters
    ----------
    dA : double
        Area of discrete integral, dx * dy

    Z : array[complex]
        2D map of size [Ny, Nx]

    Unn : array[complex]
        3D array of size [Nm, Nm, Nx]. This should contain the
        Un(x) * Un'(x)**C products

    Umm : array[complex]
        3D array of size [Nm, Nm, Ny]. This should contain the
        Um(x) * Um'(x)**C products

    tmp : array[complex]
        Temporary storage that can be used to compute the dot
        products between Z and Unn.
        Should be of size (Nm, Nm, Ny)

    result : array[complex]
        Resulting Knmnm output of size [Nm, Nm, Nm, Nm].
        IMPORTANT: Note output indexing of result K[m,a,n,b]
    """
    if not (Z.ndim == 2):
        raise RuntimeError("Z.ndim != 2")
    if not (result.ndim == 4):
        raise RuntimeError("result.ndim != 2")
    cdef:
        int Nx = Z.shape[1]
        int Ny = Z.shape[0]
        int Nm = result.shape[0]
    if not (result.shape[0] == Nm and result.shape[1] == Nm and result.shape[2] == Nm and result.shape[3] == Nm):
        raise RuntimeError("result.shape != (Nm, Nm, Nm, Nm)")
    if not (Unn_.shape[0] == Nm and Unn_.shape[1] == Nm and Unn_.shape[2] == Nx):
        raise RuntimeError("Unn_.shape != (Nm, Nm, Nx)")
    if not (Umm_.shape[0] == Nm and Umm_.shape[1] == Nm and Umm_.shape[2] == Ny):
        raise RuntimeError("Umm_.shape != (Nm, Nm, Ny)")
    if not (tmp.shape[0] == Nm and tmp.shape[1] == Nm and tmp.shape[2] == Ny):
        raise RuntimeError("tmp.shape != (Nm, Nm, Ny)")
    c_riemann_optimised(
        Nx, Ny, Nm, dA,
        &Z[0,0],
        &Unn_[0,0,0],
        &Umm_[0,0,0],
        &tmp[0,0,0],
        &result[0,0,0,0],
    )


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef void c_riemann_optimised(
            int Nx,
            int Ny,
            int Nm,
            double dA,
            complex_t* Z, # [Ny, Nx]
            complex_t* Unn_, # [Nm, Nm, Nx]
            complex_t* Umm_, # [Nm, Nm, Ny]
            complex_t* tmp,  # [Nm, Nm, Ny]
            complex_t* result, # [Nm, Nm, Nm, Nm]
        ) noexcept nogil:
    """Raw interface that can be accessed via a Python interface using
    `compute_map_scattering_coeffs_riemann_optimised`, see that function
    for more details.

    Parameters
    ----------
    Nx : int
        Number of x samples

    Ny : int
        Number of y samples

    Nm : int
        Number of 1D modes

    dA : double
        Area of discrete integral, dx * dy

    Z : array[complex]
        2D map of size [Ny, Nx]

    Unn_ : array[complex]
        3D array of size [Nm, Nm, Nx]. This should contain the
        Un(x) * Un'(x)**C products

    Umm_ : array[complex]
        3D array of size [Nm, Nm, Ny]. This should contain the
        Um(x) * Um'(x)**C products

    tmp : array[complex]
        Temporary storage that can be used to compute the dot
        products between Z and Unn_.
        Should be of size (Nm, Nm, Ny)

    result : array[complex]
        Resulting Knmnm output of size [Nm, Nm, Nm, Nm].
        IMPORTANT: Note output indexing of result K[m,a,n,b]
    """
    cdef:
        Py_ssize_t i, k
        int lda, ldb, ldc, M, N, K
        complex_t alpha = dA
        complex_t beta = 0
        complex_t *A
        complex_t *B
        complex_t *C

    # setup blas zgemm variables for map product with x-modes
    # store this in a temporary memory and then apply the
    M = Ny
    N = Nm
    K = Nx
    lda = K
    ldb = K
    ldc = M
    A = Z

    for i in range(Nm):
        B = Unn_ + i*Nm*Nx # &Unn_[i,0,0]
        C = tmp + i*Nm*Ny #&tmp[i,0,0]
        zgemm("T", "N",
            &M, &N, &K,
            &alpha, A, &lda, B, &ldb,
            &beta, C, &ldc
        )

    # Now we need to perform the product of the y-modes
    # with all the map and x-mode products
    M = Nm
    N = Nm
    K = Ny
    lda = Ny
    ldb = Ny
    ldc = Nm
    alpha = 1 # reset as already applied dA
    for i in range(Nm):
        for k in range(Nm):
            A = Umm_ + i*Nm*Ny # &Umm_[i,0,0]
            B = tmp + k*Nm*Ny # &tmp[k,0,0]
            C = result + i*Nm*Nm*Nm + k*Nm*Nm # &result[i,k,0,0]
            zgemm("T", "N",
                &M, &N, &K,
                &alpha, A, &lda, B, &ldb,
                &beta, C, &ldc
            )


def composite_newton_cotes_weights(N, order):
    """
    Constructs the weights for a composite Newton-Cotes rule for integration
    along 1-dimensional line with equally spaced steps. Newton-Cotes are a
    generalisation of a various discrete integration methods that are
    approximating an integrated by some polynomial. Common methods are:

        N = 0 : Riemann sum
        N = 1 : Trapezoid rule
        N = 2 : Simpsons rule
        N = 4 : Simpsons 3/8 rule

    Approximating a large bound with a high order polynomial can be numerically
    problematic. Thus a composite rule is generated by subdividing a larger area
    into multiple chunks and applying each rule along it.

    If the order Newton-Cotes order specified does not produce a rule that fits
    into N, the order is decremented and that is used to fill gaps that do
    no fit.

    See https://mathworld.wolfram.com/Newton-CotesFormulas.html

    Parameters
    ----------
    N : integer >= 0
        1D size of array being ingrated over

    order : integer >= 0
        Order of Newton-Cotes rule to use

    Returns
    -------
    weights : array
        Array of weights to multiply data with before summing
    """
    # ensure we have a 1xN array
    if order == 0:
        return np.ones((N,))
    W = newton_cotes(order, 1)[0]
    wx = np.zeros(N, dtype=np.float64)
    i = 0
    while i < N - 1:
        try:
            wx[i:(i+len(W))] += W
            i += len(W)-1
        except ValueError as ex:
            # if we can't fit a newton-cotes rule in to the space
            # needed then pick a slightly smaller one
            if len(wx[i:(i+len(W))]) < len(W):
                order -= 1
                W = newton_cotes(order, 1)[0]
    return wx


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
def map_coupling_matrix_riemann(complex_t[:,::1] Y, double dx, double dy, complex_t[:,::1] Un, complex_t[:,::1] Um, long[:,::1] index_map):
    cdef:
        Py_ssize_t i, j, k, l, n, m, nn, mm, Nm
        complex_t a, b
        complex_t[:, ::1] K
        double D = dx * dy
        complex_t sum = 0
        complex_t [:, ::1] Un_
        complex_t [:, ::1] Um_

    Nm = index_map.shape[0]
    Nx = Un.shape[1]
    Ny = Um.shape[1]
    K = np.zeros((Nm, Nm), dtype=np.complex128)
    Un_ = np.conj(Un)
    Um_ = np.conj(Um)

    for i in range(Nm):
        n = index_map[i, 0]
        m = index_map[i, 1]
        for j in range(Nm):
            nn = index_map[j, 0]
            mm = index_map[j, 1]
            sum = 0
            for k in range(Nx):
                for l in range(Ny):
                    a = Un[n, k] * Un_[nn, k]
                    b = Um[m, l] * Um_[mm, l]
                    sum += a * b * Y[k, l]

            K[i, j] = K[i, j]  + sum * D

    return np.asarray(K)

# WIP integration function
# needs to separate the real H(x) functions for integrand and the complex
# scaling factors still.

# @cython.boundscheck(False)
# @cython.wraparound(False)
# @cython.initializedcheck(False)
# cpdef void compute_map_knm_scatterins_riemann_optimised_mode_matched(
#     double dA,
#     double[:,::1] Z_real,
#     double[:,::1] Z_imag,
#     double[:,:,::1] Unn_,
#     double[:,:,::1] Umm_,
#     double[:,:,::1] tmp_real,
#     double[:,:,::1] tmp_imag,
#     double[:,:,:,::1] result_real,
#     double[:,:,:,::1] result_imag,
#     complex_t[:,:,:,::1] result,
# ) except *:
#     cdef:
#         Py_ssize_t i, j, k, l,
#         int Nx, Ny, Nm, inc=1, lda, ldb, ldc, M, N, K
#         double alpha = dA
#         double beta = 0
#         double *A
#         double *B
#         double *C
#         double *X
#         double *Y

#     Nx = Z_real.shape[1]
#     Ny = Z_real.shape[0]
#     Nm = result_real.shape[0] # computing (Nm, Nm, Nm, Nm) output
#     assert(result_real.shape[0] == result_real.shape[1] == result_real.shape[2] == result_real.shape[3])
#     assert(result_imag.shape[0] == result_imag.shape[1] == result_imag.shape[2] == result_imag.shape[3])
#     assert(Unn_.shape[0] == Unn_.shape[1] == result_imag.shape[0] == result_real.shape[0])
#     assert(Unn_.shape[0] == Unn_.shape[1] == result_imag.shape[0] == result_real.shape[0])
#     assert(Unn_.shape[2] == Nx)
#     assert(Umm_.shape[2] == Ny)

#     # setup blas zgemm variables for map product with x-modes
#     # store this in a temporary memory and then apply the
#     M = Ny
#     N = Nm
#     K = Nx
#     lda = K
#     ldb = K
#     ldc = M
#     for i in range(Nm):
#         B = &Unn_[i,0,0]

#         A = &Z_real[0, 0]
#         C = &tmp_real[i,0,0]
#         dgemm("T", "N",
#             &M, &N, &K,
#             &alpha, A, &lda, B, &ldb,
#             &beta, C, &ldc
#         )
#         A = &Z_imag[0, 0]
#         C = &tmp_imag[i,0,0]
#         dgemm("T", "N",
#             &M, &N, &K,
#             &alpha, A, &lda, B, &ldb,
#             &beta, C, &ldc
#         )

#     # Now we need to perform the product of the y-modes
#     # with all the map and x-mode products
#     M = Nm
#     N = Nm
#     K = Ny
#     lda = Ny
#     ldb = Ny
#     ldc = Nm
#     alpha = 1 # reset as already applied dA
#     for i in range(Nm):
#         for k in range(Nm):
#             A = &Umm_[i,0,0]
#             B = &tmp_real[k,0,0]
#             C = &result_real[i,k,0,0]
#             dgemm("T", "N",
#                 &M, &N, &K,
#                 &alpha, A, &lda, B, &ldb,
#                 &beta, C, &ldc
#             )
#             B = &tmp_imag[k,0,0]
#             C = &result_imag[i,k,0,0]
#             dgemm("T", "N",
#                 &M, &N, &K,
#                 &alpha, A, &lda, B, &ldb,
#                 &beta, C, &ldc
#             )

#     for i in range(Nm):
#         for j in range(Nm):
#             for k in range(Nm):
#                 for l in range(Nm):
#                     result[i,j,k,l].real = result_real[i,j,k,l]

#     for i in range(Nm):
#         for j in range(Nm):
#             for k in range(Nm):
#                 for l in range(Nm):
#                     result[i,j,k,l].imag = result_imag[i,j,k,l]



"""A bunch of tests for serialisation of models.

Writing tests for every combination might be a bit overkill, so we'll just test a few
different complicated cases and probably copy issues in when we find them.
"""

import finesse
import finesse.components as fc
import finesse.detectors as fd
import numpy as np
import pytest
from pathlib import Path
from finesse.utilities.maps import circular_aperture


@pytest.fixture(params=[True, False])
def model(request):
    use_map = request.param

    model = finesse.Model()
    model.add_parameter("test", 2245)
    model.modes("even", maxtem=2)
    LASER = model.add(fc.Laser("LASER", P=1))
    PRM = model.add(fc.Mirror("PRM", T=0.031, L=0, Rc=-10.948))
    PR2 = model.add(fc.Beamsplitter("PR2", T=230.0e-6, L=0, Rc=-4.543, alpha=0))
    PR3 = model.add(fc.Beamsplitter("PR3", T=4.0e-6, L=0, Rc=36.021, alpha=0))
    LENS = model.add(fc.Lens("LENS", f=50e3))
    ITM = model.add(fc.Mirror("ITM", T=0.015, L=0, Rc=1940))
    ETM = model.add(fc.Mirror("ETM", T=0, L=75e-6, Rc=model.test.ref))
    model.add(fc.ReadoutDC("DC", optical_node=PRM.p2.o))

    model.link(
        LASER.p1,
        PRM.p2,
        PRM.p1,
        16608.6e-3,
        PR2.p1,
        PR2.p2,
        16093e-3,
        PR3.p1,
        PR3.p2,
        19537.4e-3 + 4829.6e-3 + 40.0e-3,
        LENS,
        ITM.p2,
        ITM.p1,
        3994.485,
        ETM.p1,
    )

    model.add(fc.Cavity("ARM", ITM.p1.o))
    model.add(fc.Cavity("PRC", ITM.p2.o, via=PRM.p1.o))

    model.add(fd.PowerDetector("P_in", PRM.p2.i))
    model.add(fd.PowerDetector("P_prc", ITM.p2.i))
    model.add(fd.PowerDetector("P_itm", ITM.p1.o))
    model.add(fd.FieldDetector("E_itm", ITM.p1.i, f=0))
    model.add(fd.PowerDetector("P_etm", ETM.p1.i))
    model.add(fd.BeamPropertyDetector("q_in", PRM.p2.i, "q"))
    model.add(fd.BeamPropertyDetector("q_prc", ITM.p2.i, "q"))
    model.add(fd.BeamPropertyDetector("q_itm", ITM.p1.o, "q"))
    model.add(fd.BeamPropertyDetector("q_etm", ETM.p1.o, "q"))

    if use_map:
        x = np.linspace(-0.16, 0.16, 100)
        y = np.linspace(-0.16, 0.16, 101)
        ITM.surface_map = finesse.knm.Map(x, y, amplitude=circular_aperture(x, y, 0.16))

    model.add(
        fd.FieldScanLine(
            "beam_itm",
            ITM.p1.o,
            100,
            xlim=[-0.16, 0.16],
            w0_scaled=False,
        )
    )
    model.add(
        fd.FieldScanLine(
            "beam_prc",
            ITM.p2.i,
            100,
            xlim=[-0.16, 0.16],
            w0_scaled=False,
        )
    )
    model.add(fd.MathDetector("Rc_ITM", ITM.Rcx.ref))
    model.add(fd.MathDetector("Rc_ETM", ETM.Rcx.ref))
    model.add(fd.MathDetector("f_ITM", LENS.f.ref))

    model.parse(
        """
    mathd PRG P_prc/P_in
    mathd ARG P_itm/P_prc
    # rescale to rough unitary values for optimisation as we know what the arm and
    # prc gains should be in the ideal case
    mathd cost ARG/260+PRG/50
    """
    )
    return model


@pytest.fixture
def test_save_load(model, tmpdir: Path):
    filepath: Path = tmpdir + "model.pkl"
    model.save(filepath)
    loaded = finesse.model.load(filepath)
    return model.deepcopy(), loaded.deepcopy()


def test_coupled_cavity_homs(test_save_load):
    model, loaded = test_save_load

    for a in model.all_parameters:
        b = loaded.get(a)
        assert a is not b
        if a.is_symbolic or b.is_symbolic:
            assert str(a.value) == str(b.value)
        else:
            assert a.value == b.value
        assert a.name == b.name
        assert a.__class__ == b.__class__

    for a, b in zip(loaded.elements.values(), model.elements.values()):
        assert a is not b
        assert a.name == b.name
        assert a.__class__ == b.__class__

    model.ITM.Rc = 1980
    loaded.ITM.Rc = 1980
    assert np.all(model.ITM.Rc == loaded.ITM.Rc)
    assert np.all(loaded.ARM.g == model.ARM.g)
    assert np.all(loaded.PRC.g == model.PRC.g)


def test_noxaxis(test_save_load):
    model, loaded = test_save_load

    A = model.run()
    B = loaded.run()

    for output in A.outputs:
        assert np.allclose(A[output], B[output], rtol=1e-10), output


@pytest.mark.parametrize(
    "action",
    [
        "xaxis(ITM.Rcx, lin, -10, 10, 2, True)",  # change geometry, makes sure ABCDs are all updating correctly
        "xaxis(test, lin, -10, 10, 2, True)",  # change model var
        "xaxis(LASER.P, lin, 0, 1, 2, True)",  # change element var
    ],
)
def test_xaxis(test_save_load, action):
    model, loaded = test_save_load
    A = model.run(action)
    B = loaded.run(action)

    for output in A.outputs:
        if A[output].dtype == object:
            assert np.allclose(
                A[output].astype(complex), B[output].astype(complex), rtol=1e-14
            ), output
        else:
            assert np.allclose(A[output], B[output], rtol=1e-14), output


def test_dcfields(test_save_load):
    model, loaded = test_save_load
    action = "dc_fields()"
    A = model.run(action)
    B = loaded.run(action)

    assert np.allclose(A.fields, B.fields, rtol=1e-14)


def test_frequency_response(test_save_load):
    model, loaded = test_save_load
    action = "frequency_response([1, 10, 100], ITM.mech.z, ETM.mech.F_z)"
    model.fsig.f = loaded.fsig.f = 1
    A = model.run(action)
    B = loaded.run(action)
    assert np.allclose(A.out, B.out, rtol=1e-14)



"""TraceDependency ordering unit tests."""

import pytest
from finesse import Model
import finesse.components as components


@pytest.fixture
def michelson_model():
    model = Model()
    model.parse(
        """
        l L0 P=1
        s s0 L0.p1 BS.p1 L=10

        bs BS

        s sY BS.p2 ITMY.p1 L=100

        m ITMY R=0.99 T=0.01 Rc=-5580
        s LY ITMY.p2 ETMY.p1 L=10k
        m ETMY R=0.99 T=0.01 Rc=5580

        s sX BS.p3 ITMX.p1 L=100

        m ITMX R=0.99 T=0.01 Rc=-5580
        s LX ITMX.p2 ETMX.p1 L=10k
        m ETMX R=0.99 T=0.01 Rc=5580
        """
    )
    return model


### Only cavities ###


def test_implicit_arm_cavity_ordering__1(michelson_model):
    """Test that cavXARM comes before cavYARM via alphabetic ordering."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    assert michelson_model.trace_order == [FPX, FPY]


def test_implicit_arm_cavity_ordering__2(michelson_model):
    """Test that adding cavYARM cavity before cavXARM makes no difference to order."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2)
    michelson_model.add(FPY)
    michelson_model.add(FPX)

    assert michelson_model.trace_order == [FPX, FPY]


def test_explicit_arm_cavity_ordering__x_priority(michelson_model):
    """Test that giving cavXARM priority arg puts it first."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=1)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    assert michelson_model.trace_order == [FPX, FPY]


def test_explicit_arm_cavity_ordering__y_priority(michelson_model):
    """Test that giving cavYARM priority arg puts it first."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=1)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    assert michelson_model.trace_order == [FPY, FPX]


def test_explicit_arm_cavity_ordering__x_priority_last(michelson_model):
    """Test that giving cavXARM priority arg puts it first."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=-1)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    assert michelson_model.trace_order == [FPY, FPX]


def test_explicit_arm_cavity_ordering__xy_priority(michelson_model):
    """Test that giving cavXARM and cavYARM priority args with X having higher
    priority puts it first."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=2)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=1)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    assert michelson_model.trace_order == [FPX, FPY]


def test_explicit_arm_cavity_ordering__xy_priority_last(michelson_model):
    """Test that giving cavXARM and cavYARM negative priority args with X having higher
    priority puts it first."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=-1)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=-2)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    assert michelson_model.trace_order == [FPX, FPY]


def test_explicit_arm_cavity_ordering__change_priority(michelson_model):
    """Test that changing cavYARM priority to highest, after adding the cavs
    to the model, results in it being first in the trace order."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=2)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=1)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    FPY.priority = 3

    assert michelson_model.trace_order == [FPY, FPX]


def test_explicit_arm_cavity_ordering__change_priority_first_to_last(michelson_model):
    """Test that changing cavYARM priority to highest, after adding the cavs
    to the model, results in it being first in the trace order."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=1)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=2)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    FPY.priority = -3

    assert michelson_model.trace_order == [FPX, FPY]


def test_explicit_arm_cavity_ordering__change_priority_last_to_first(michelson_model):
    """Test that changing cavYARM priority to highest, after adding the cavs
    to the model, results in it being first in the trace order."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=1)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=-5)
    michelson_model.add(FPX)
    michelson_model.add(FPY)

    FPY.priority = 2

    assert michelson_model.trace_order == [FPY, FPX]


def test_explicit_arm_cavity_ordering__xy_equal_priority(michelson_model):
    """Test that giving cavXARM and cavYARM equal priority args puts X first."""
    FPX = components.Cavity("cavXARM", michelson_model.ITMX.p2, priority=1)
    FPY = components.Cavity("cavYARM", michelson_model.ITMY.p2, priority=1)
    michelson_model.add(FPY)
    michelson_model.add(FPX)

    assert michelson_model.trace_order == [FPX, FPY]


### Only gausses ###


def test_implicit_gauss_ordering__1(michelson_model):
    """Test that gAS comes before gL0 via alphabetic ordering."""
    gL0 = components.Gauss("gL0", michelson_model.L0.p1.o, q=-10 + 0.1j)
    gAS = components.Gauss("gAS", michelson_model.BS.p4.o, q=0 + 0.15j)
    michelson_model.add(gAS)
    michelson_model.add(gL0)

    assert michelson_model.trace_order == [gAS, gL0]


def test_implicit_gauss_ordering__2(michelson_model):
    """Test that adding gL0 before gAS makes no difference to order."""
    gL0 = components.Gauss("gL0", michelson_model.L0.p1.o, q=-10 + 0.1j)
    gAS = components.Gauss("gAS", michelson_model.BS.p4.o, q=0 + 0.15j)
    michelson_model.add(gL0)
    michelson_model.add(gAS)

    assert michelson_model.trace_order == [gAS, gL0]


def test_explicit_gauss_ordering__L0_priority(michelson_model):
    """Test that giving gL0 priority arg puts it first."""
    gL0 = components.Gauss("gL0", michelson_model.L0.p1.o, q=-10 + 0.1j, priority=1)
    gAS = components.Gauss("gAS", michelson_model.BS.p4.o, q=0 + 0.15j)
    michelson_model.add(gAS)
    michelson_model.add(gL0)

    assert michelson_model.trace_order == [gL0, gAS]


def test_explicit_gauss_ordering__both_priority(michelson_model):
    """Test that giving gL0 and gAS priority args with gL0 having higher
    priority puts it first."""
    gL0 = components.Gauss("gL0", michelson_model.L0.p1.o, q=-10 + 0.1j, priority=1)
    gAS = components.Gauss("gAS", michelson_model.BS.p4.o, q=0 + 0.15j, priority=0.5)
    michelson_model.add(gAS)
    michelson_model.add(gL0)

    assert michelson_model.trace_order == [gL0, gAS]


def test_explicit_gauss_ordering__change_priority(michelson_model):
    """Test that changing gL0 priority to lowest, after adding the gausses
    to the model, results in it being last in the trace order."""
    gL0 = components.Gauss("gL0", michelson_model.L0.p1.o, q=-10 + 0.1j, priority=2)
    gAS = components.Gauss("gAS", michelson_model.BS.p4.o, q=0 + 0.15j, priority=1)
    michelson_model.add(gL0)
    michelson_model.add(gAS)

    gL0.priority = 0.5

    assert michelson_model.trace_order == [gAS, gL0]


def test_explicit_gauss_ordering__both_equal_priority(michelson_model):
    """Test that giving gL0 and gAS equal priority args puts gAS first."""
    gL0 = components.Gauss("gL0", michelson_model.L0.p1.o, q=-10 + 0.1j, priority=2.5)
    gAS = components.Gauss("gAS", michelson_model.BS.p4.o, q=0 + 0.15j, priority=2.5)
    michelson_model.add(gL0)
    michelson_model.add(gAS)

    assert michelson_model.trace_order == [gAS, gL0]


# TODO (sjr) Write some tests with multiple mixed dependencies



import pytest
from finesse.script import parse
from finesse.script.exceptions import KatScriptError
from testutils.text import dedent_multiline, escape_full


@pytest.mark.parametrize(
    "script",
    (
        # Positional argument.
        "lambda(1550e-9,)",
        "lambda(1550e-9 ,)",
        "lambda(1550e-9, )",
        "lambda(1550e-9 , )",
        # Keyword argument.
        "modes(maxtem=3,)",
        "modes(maxtem=3 ,)",
        "modes(maxtem=3, )",
        "modes(maxtem=3 , )",
        # Mixed arguments.
        "modes(odd, maxtem=7,)",
        "modes(odd, maxtem=7 ,)",
        "modes(odd, maxtem=7, )",
        "modes(odd, maxtem=7 , )",
    ),
)
def test_trailing_comma(script):
    parse(script)


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            dedent_multiline(
                """
                modes(maxtem=1)
                modes(maxtem=2)
                modes(maxtem=3)
                """
            ),
            (
                "\nlines 1-3: there can only be one 'modes' directive\n"
                "-->1: modes(maxtem=1)\n"
                "      ^^^^^\n"
                "-->2: modes(maxtem=2)\n"
                "      ^^^^^\n"
                "-->3: modes(maxtem=3)\n"
                "      ^^^^^"
            ),
            id="modes",
        ),
    ),
)
def test_duplicate_singular_functions_invalid(script, error):
    with pytest.raises(KatScriptError, match=escape_full(error)):
        parse(script)


@pytest.mark.parametrize(
    "script",
    (
        pytest.param(
            dedent_multiline(
                """
                l l1
                m m1
                m m2
                link(l1, m1)
                link(m1, m2)
                """
            )
        ),
        pytest.param(
            dedent_multiline(
                """
                l l1
                readout_dc pd1
                butter F 4 lowpass 100
                amplifier G 1
                amplifier K 2
                amplifier L 3

                link(l1, pd1, pd1.DC, F, G, l1.amp)
                link(G.p2, K, F.p1)
                link(G.p2, l1.frq)
                link(K.p2, L, l1.frq)
                """
            )
        ),
    ),
)
def test_duplicate_nonsingular_functions_valid(script):
    """Test that non-singular instructions can be specified multiple times."""
    parse(script)



# cython: profile=False
import logging

from finesse.detectors.general import NoiseDetector
from finesse.components.general import NoiseType
from finesse.simulations.simulation cimport BaseSimulation
from finesse.simulations.sparse.solver cimport SparseSolver
from finesse.components.modal.workspace cimport KnmConnectorWorkspace
from finesse.components.workspace cimport ConnectorCallbacks
from finesse.components.workspace cimport ConnectorWorkspace, ConnectorMatrixSimulationInfo
from finesse.cymath.cmatrix cimport KLUMatrix

LOGGER = logging.getLogger(__name__)


cdef class SparseMatrixSimulation(BaseSimulation):
    """
    A simulation type which solves the carrier and signal simulations using large
    sparse matrices, solving systems such as Ma=b.
    """

    def __init__(self, model, unicode name, dict simulation_options, bint needs_matrix=True):
        super().__init__(model, name, simulation_options)

    def build_carrier_solver(self, nodes, optical_frequencies):
        if not issubclass(self.simulation_options["carrier_solver"], SparseSolver):
            raise Exception("Solver should inherit from SparseSolver")
        return self.simulation_options["carrier_solver"](
            "carrier",
            nodes,
            optical_frequencies,
            None,
            False,
            self.model.force_refill,
            {},
            self.model_settings.num_HOMs,
            debug_mode=self.simulation_options["debug_mode"],
            matrix_kwargs=self.simulation_options.get("carrier_solver_kwargs", {})
        )

    def build_signal_solver(self, nodes, optical_frequencies, signal_frequencies):
        if not issubclass(self.simulation_options["signal_solver"], SparseSolver):
            raise Exception("Solver should inherit from SparseSolver")

        return self.simulation_options["signal_solver"](
            "signal",
            nodes,
            optical_frequencies,
            signal_frequencies,
            True,
            self.model.force_refill,
            {},
            self.model_settings.num_HOMs,
            matrix_kwargs=self.simulation_options.get("signal_solver_kwargs", {})
        )

    def setup_build(self):
        if self.compute_signals:
            self.initialise_noise_matrices()

        self.initialise_workspaces()
        self.update_all_parameter_values()

        if self.compute_signals:
            self.initialise_noise_sources()
            self.initialise_noise_selection_vectors()

        if self.is_modal:
            # compute all the initial:
            #     - scattering matrices
            #     - space Gouy phases
            #     - laser tem Gouy phases
            for ws in self.workspaces:
                # if the connector scatters modes then initialise the
                # knm workspaces here and store the connector workspace
                # in to_scatter_matrix_compute for future use
                if isinstance(ws, (KnmConnectorWorkspace)):
                    (<KnmConnectorWorkspace> ws).initialise_knm_workspaces()
                    self.to_scatter_matrix_compute.append(ws)
            self.compute_knm_matrices()
            self.set_gouy_phases()
            # ... then determine which beam parameters will be changing
            # so that only the items from above which change get
            # re-computed on subsequent calls to their respective functions
            self._determine_changing_beam_params()

        self.carrier.initialise(self)
        self.carrier.assign_operators(self.workspaces)
        self.carrier.construct()
        self.carrier.initial_run()

        if self.signal:
            self.signal.initialise(self)
            self.signal.assign_operators(self.workspaces)
            self.signal.assign_noise_operators(self.workspaces)
            self.signal.construct()
            self.signal.initial_run()

    cpdef modal_update(self):
        if not BaseSimulation.modal_update(self):
            return False
        # Compute the changing scattering matrices
        self.compute_knm_matrices()
        return True

    cpdef compute_knm_matrices(self):
        cdef KnmConnectorWorkspace ws
        for ws in self.to_scatter_matrix_compute:
            ws.update_changing_knm_workspaces()
            ws.compute_scattering_matrices()
            # TODO (sjr) Probably want a flag to check if quantum noise calcs
            #            being performed and to only do this call if so
            ws.compute_knm_losses()

    cpdef initialise_workspaces(self) :
        cdef ConnectorMatrixSimulationInfo info
        from finesse.components import Connector, Cavity
        # TODO ddb - probably need to move away from lists as they aren't that fast to iterate
        # over. Maybe once we have all the lists filled we can covert them into some PyObject
        # memoryview
        self.workspace_name_map = {}
        self.workspaces = []
        self.cavity_workspaces = {}
        self.to_scatter_matrix_compute = []
        self.gouy_phase_workspaces = []

        if self.is_modal and self.trace == NULL:
            raise Exception("Beam trace has not been set before workspaces are initialised")

        # Get any callbacks for the elements in the model
        # tell the element that we have now built the model and it
        # should do some initialisations for running simulations
        for el in self.model.elements.values():
            el._setup_changing_params()

            if isinstance(el, Connector):
                ws = el._get_workspace(self)
                if ws is None:
                    continue

                self.workspaces.append(ws) # store all workspaces here
                self.workspace_name_map[el.name] = ws

                if isinstance(ws, ConnectorWorkspace):
                    # Determine if we should be adding this workspace to any
                    # todo list for looping over later

                    # Here we grab the information objects from each workspace
                    # which describes the connections and filling to be done
                    # for each element for each simulation type
                    if self.signal:
                        x = (
                            (ws.carrier, self.carrier),
                            (ws.signal, self.signal)
                        )
                    else:
                        x = ((ws.carrier, self.carrier),)

                    for info, mtx in x:
                        ws_store = mtx.workspaces
                        if info.callback_flag & ConnectorCallbacks.FILL_MATRIX:
                            ws_store.to_initial_fill.append(ws) # Initial fill all
                            if info.matrix_fills.num_refills > 0 or mtx.forced_refill:
                                ws_store.to_refill.append(ws)

                        if info.callback_flag & ConnectorCallbacks.FILL_RHS:
                            ws_store.to_rhs_refill.append(ws)

                        if info.callback_flag & ConnectorCallbacks.FILL_NOISE:
                            try:
                                if any([x in self.signal.noise_sources for x in ws.owner.noises]):
                                    ws_store.to_noise_refill.append(ws)
                            except AttributeError:
                                # Component isn't a NoiseGenerator, but can still generate quantum
                                # noise
                                if NoiseType.QUANTUM in self.signal.noise_sources:
                                    ws_store.to_noise_refill.append(ws)

                        # Quantum noise is special, as all connectors can be a source if they have
                        # open inputs
                        if info.callback_flag & ConnectorCallbacks.FILL_INPUT_NOISE:
                            if NoiseType.QUANTUM in self.signal.noise_sources:
                                ws_store.to_noise_input_refill.append(ws)

                    if ws.fn_gouy_c is not None or ws.fn_gouy_py is not None:
                        self.gouy_phase_workspaces.append(ws)

                elif ws is not None:
                    # None means the component doesn't want anything
                    # to do with this simulation
                    raise Exception("Unexpected workspace type")
            elif isinstance(el, Cavity):
                self.cavity_workspaces[el] = el._get_workspace(self)
                self.workspace_name_map[el.name] = self.cavity_workspaces[el]

        # Compile cy_exprs for changing symbolics, these are stored
        # in ElementWorkspace.chprm_expr which is used for fast evaluating
        # of the changing symbolic expressions
        for ws in self.workspaces:
            ws.compile_cy_exprs()
            # Also compile the changing ABCD matrix elements, these are
            # stored in the relevant cy_expr** field of the associated
            # workspace -> note that the cy_expr* element is NULL for non
            # changing elements
            if self.is_modal:
                ws.compile_abcd_cy_exprs()

        LOGGER.info("Refilling carrier elements:\n%s", self.carrier.workspaces.to_refill)

        self.carrier.workspaces.list_to_C()
        if self.signal:
            LOGGER.info("Refilling signal elements:\n%s", self.signal.workspaces.to_refill)
            self.signal.workspaces.list_to_C()
        #print(self.name)
        #print("MATRIX")
        #print("carrier to_refill", len(self.carrier.workspaces.to_refill))
        #print("signal to_refill", len(self.signal.workspaces.to_refill))
        #print("to_rhs_refill", self.to_rhs_refill)


    cpdef initialise_noise_matrices(self) :
        from finesse.detectors.general import NoiseDetector

        # Which noise types are we measuring?
        for el in self.model.detectors:
            if isinstance(el, NoiseDetector):
                if el.noise_type not in self.signal.noise_sources:
                    self.signal.noise_sources[el.noise_type] = []
                    self.signal.add_noise_matrix(el.noise_type)

    cpdef initialise_noise_sources(self) :
        from finesse.components.general import NoiseGenerator

        for el in self.model.elements.values():
            if isinstance(el, NoiseGenerator):
                for _type, nodes in el.noises.items():
                    # Only consider noise sources for the types of noise we'll be measuring
                    if _type in self.signal.noise_sources:
                        self.signal.noise_sources[_type].append((el, nodes))

    cpdef initialise_noise_selection_vectors(self) :
        cdef KLUMatrix mtx = self.signal._M
        for el in self.model.elements.values():
            if hasattr(el, "_requested_selection_vectors"):
                for name in el._requested_selection_vectors:
                    el._requested_selection_vectors[name] = mtx.request_rhs_view()



"""Objects for connecting and registering connections between components."""

from __future__ import annotations

import enum
import logging
import weakref
from collections import OrderedDict
from copy import deepcopy
from typing import TYPE_CHECKING, Callable

from .. import components
from ..env import warn
from ..exceptions import ComponentNotConnected, FinesseException
from ..freeze import Freezable
from ..utilities import check_name, is_iterable, OrderedSet
from ..symbols import Constant

if TYPE_CHECKING:
    from finesse.element import ModelElement
    from finesse.detectors import Detector
    from finesse.components.general import MechanicalConnector

LOGGER = logging.getLogger(__name__)


@enum.unique
class NodeType(enum.Enum):
    """Enum describing the physical connection type of a :class:`.Node`"""

    OPTICAL = 0
    ELECTRICAL = 1
    MECHANICAL = 2


@enum.unique
class NodeDirection(enum.Enum):
    """Enum describing the direction that information at a :class:`.Node` flows.

    This is largely a description to help understand how external information flows in
    and out of a component. Inside a component all nodes will couple to one another in
    more complex ways.

    Input nodes are those going into a component, whereas output describe those leaving.
    For example incident and reflected light fields.

    Bidrectional takes information either direction. For example a mechanical degree of
    freedom, external forces can be applied to it, or its motion can be coupled to some
    external system.
    """

    INPUT = 0
    OUTPUT = 1
    BIDIRECTIONAL = 2


class Port(Freezable):
    """A collection of all the nodes at a specific point/surface of a component.

    Parameters
    ----------
    name : str
        Name of newly created node.

    component : Sub-class of :class:`.Connector`
        The component that this node belongs to.

    node_type : :class:`.NodeType`
        Physical node type.
    """

    def __init__(self, name, component, node_type):
        self._unfreeze()
        self.__component = weakref.ref(component)
        self.__name = check_name(name)
        self.__full_name = "{}.{}".format(component.name, name)
        self.__type = node_type
        self.__nodes = OrderedDict()
        self.__enabled = True
        # keep track of Mechanical connectors already using this port to prevent
        # double connections
        self.mechanical_connection: MechanicalConnector | None = None
        self._freeze()

    def __deepcopy__(self, memo):
        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))
        # Manually update the weakrefs to be correct
        new.__component = weakref.ref(memo[id(self.component)])
        return new

    def _add_node(self, name, direction=None, node=None, **kwargs):
        """Adds a new node to this port. Once called the new node can be access with:

            obj.port.name

        Parameters
        ----------
        name : str
            Name of the new node
        dof_parameter : ModelParameter
            A model parameter to be associated with this node

        Returns
        -------
        Node object added
        """
        if name in self.__nodes:
            raise Exception("Node already added to this component")

        if node is None:
            if self.__type == NodeType.OPTICAL:
                if direction is None:
                    raise Exception(
                        "Node direction must be specified for optical nodes"
                    )
                node = OpticalNode(name, self, direction, **kwargs)
            elif (
                self.__type == NodeType.MECHANICAL or self.__type == NodeType.ELECTRICAL
            ):
                node = SignalNode(name, self, direction, self.__type, **kwargs)
            else:
                raise Exception("Unexpected node type")
        elif len(kwargs) != 0:
            raise Exception(f"Cannot use kwargs when node ({node!r}) is not None")

        if node is None:
            raise RuntimeError("Node unexpectedly None")

        if node.type == NodeType.OPTICAL or self.type == NodeType.OPTICAL:
            if node.type != self.type:
                raise Exception(
                    f"Node ({node.type}) and port type ({self.type}) must be the same ({node!r}, {self!r})"
                )

        self.__nodes[name] = node
        # Update Elements node dict so it doesn't have to keep checking
        # it's ports about what nodes it has
        self.__component()._Connector__nodes[node.full_name] = node
        self._unfreeze()
        assert not hasattr(self, name)
        setattr(self, name, node)
        self._freeze()
        return node

    def _replace_node(self, name, new_node):
        """Replaces a node at this port with a new one.

        This can be used to change connections between elements but care must be taken.
        This does not update any states of the elements beyond simply changing the node.
        This should only be used when an element is borrowing a node from another
        element, and you want to change which node it is borrowing.
        """
        old_node = self.__nodes[name]
        if name not in self.__nodes:
            raise FinesseException(f"{self!r} does not have a node called `{name}`")
        if new_node.type != self.type:
            raise FinesseException("Node and port type must be the same")
        self.__nodes[name] = new_node
        # Update Elements node dict so it doesn't have to keep checking
        # it's ports about what nodes it has
        del self.__component()._Connector__nodes[old_node.full_name]
        self.__component()._Connector__nodes[new_node.full_name] = new_node
        self._unfreeze()
        setattr(self, name, new_node)
        self._freeze()

    def __repr__(self):
        return f"<Port {self.component.name}.{self.name} Type={self.__type} @ {hex(id(self))}>"

    @property
    def full_name(self):
        """:`getter`: Returns a full name of the port: {component name}.{port name}"""
        return self.__full_name

    @property
    def enabled(self):
        return self.__enabled

    @property
    def type(self):
        """:class:`.NodeType` of the port object.

        :`getter`: Returns the node-type of the port (read-only).
        """
        return self.__type

    @property
    def name(self):
        """Name of the port object.

        :`getter`: Returns the name of the port (read-only).
        """
        return self.__name

    @property
    def _model(self):
        return self.component._model

    @property
    def component(self) -> ModelElement:
        """The component which has ownership of this port.

        :`getter`: Returns the component that this port belongs to (read-only).
        """
        return self.__component()

    @property
    def is_connected(self):
        """Flag indicating whether the port is attached to another component.

        :`getter`: Returns true if this port is attached to another component (read-only).
        """
        return any(len(n.connections) > 0 for n in self.nodes)

    @property
    def attached_to(self):
        """Components that this port is attached to. Optical ports are only ever
        connected to :class:`.Space` elements. Ports containing signal nodes can have
        multiple connections and returns a Set.

        :`getter`: Returns the component this port is attached to, or returns None
                   if no such connected component exists. Signal ports return a Set
                   of components attached (read-only).
        """
        if self.type == NodeType.OPTICAL:
            spaces = [_.space for _ in self.nodes]
            if all([spaces[0] == _ for _ in spaces]):
                return spaces[0]
            else:
                raise Exception(
                    "Nodes are somehow connected to different "
                    "spaces which should not happen"
                )
        else:
            attached_to = OrderedSet()
            for n in self.nodes:
                for c in n.connections:
                    attached_to.add(c)
            return attached_to

    @property
    def refractive_index(self):
        """If the port is an Optical port, this will return a symbolic value for the
        refractive index at this port. The refractive index is set by the `Space`
        elements that are attached to it.

        Returns
        -------
        nr : Symbol
            Symbolic value for refractive index
        """
        if self.type != NodeType.OPTICAL:
            raise Exception("Port type is not optical, cannot get refractive index")
        if self.attached_to:
            return self.attached_to.nr.ref
        else:
            return Constant(1)

    @property
    def space(self):
        """Space that the port is attached to. Equivalent to :attr:`Port.attached_to`.

        :`getter`: Returns the space that this port is attached to (read-only).
        """
        if self.type != NodeType.OPTICAL:
            raise Exception("Port type is not optical, cannot retrieve attached space.")
        return self.attached_to

    @property
    def nodes(self):
        """Nodes associated with the port.

        :`getter`: Returns a tuple of the associated nodes at this port (read-only).
        """
        return tuple(self.__nodes.values())

    def node(self, name):
        """Get a node at this port by its name."""
        return self.__nodes[name]

    def get_unique_node(self, predicate: Callable[[Node], bool]):
        """Returns the unique node at this port that satisfies the provided predicate.
        If multiple nodes satisfy this predicate then a RuntimeError is raised.

        Parameters
        ----------
        predicate : Callable[[Node], bool]
            A callable that accepts a Node and returns a boolean value

        Examples
        --------
        Selecting a unique output node:
            port.get_unique_node(lambda node: not node.is_input)
        """
        is_node = tuple(predicate(_) for _ in self.nodes)
        if is_node.count(True) == 1:
            idx = is_node.index(True)
            return self.nodes[idx]
        else:
            raise RuntimeError(
                f"Port {repr(self)} does not have a single node that satisfies the predicate"
            )


class Node:
    """Represents a specific connection at a component.

    Mathematically a node represents a single equation in the interferometer matrix.

    A node can only be owned by a single component instance - with weak references stored by the
    connected components.

    Parameters
    ----------
    name : str
        Name of newly created node.

    component : :class:`.Port`
        The port that this node belongs to.

    node_type : :class:`.NodeType`
        Physical node type.
    """

    def __init__(self, name, port, node_type, direction):
        self.__port = weakref.ref(port)
        # self.__component = weakref.ref(port.component)
        self.__name = check_name(name)
        self.__type = node_type
        self.__direction = direction
        self.__full_name = "{}.{}.{}".format(port.component.name, port.name, name)
        self.__port_name = "{}.{}".format(port.name, name)
        self.__tag_name = None
        self.__connection = None
        self.used_in_detector_output: list[Detector] = []
        self.has_signal_injection = False

    def __deepcopy__(self, memo):
        new = object.__new__(type(self))
        memo[id(self)] = new
        new.__dict__.update(deepcopy(self.__dict__, memo))
        id_port = id(self.port)
        # Manually update the weakrefs to be correct
        if id_port in memo:
            new.__port = weakref.ref(memo[id_port])
        else:
            # We need to update this reference later on
            # This will be called when the port property
            # is accessed. When this happens we'll peak back
            # at the memo once it has been filled and get
            # the new port reference. After this the refcount
            # for this function should goto zero and be garbage
            # collected
            def update_later():
                new.__port = weakref.ref(memo[id_port])

            new.__port = update_later  # just in case something calls
            # this weakref in the meantime
            memo[id(self._model)].after_deepcopy.append(update_later)

        # new.__component = weakref.ref(memo[id(self.component)])
        return new

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.full_name} @ {hex(id(self))}>"

    @property
    def is_input(self):
        """Flag indicating whether this node is an input to the associated component.

        :`getter`: Returns `True` if the field at this node
                 goes into `self.component` (read-only).
        """
        return self.direction == NodeDirection.INPUT

    @property
    def full_name(self):
        """Full name.

        :`getter`: Returns a full name of the node: {component name}.{port name}.{node name}
        """
        return self.__full_name

    @property
    def port_name(self):
        """Port name.

        :`getter`: Returns a shortened name of the node: {port name}.{node name}
        """
        return self.__port_name

    @property
    def type(self):
        """:class:`.NodeType` of the node object.

        :`getter`: Returns the node-type of the node (read-only).
        """
        return self.__type

    @property
    def direction(self):
        """:class:`.NodeDirection` of this node.

        This is largely a description to help understand how external
        information flow in and out of a component. Inside a component all nodes
        will couple to one another in more complex ways.

        Input nodes are those going into a component, whereas output
        describe those leaving. For example incident and reflected light fields.

        Bidrectional takes information either direction. For example a mechanical
        degree of freedom, external forces can be applied to it, or
        its motion can be coupled to some external system.

        :`getter`: Returns the directionality of the node (read-only).
        """
        return self.__direction

    @property
    def port(self) -> Port:
        """:class:`.Port` this node is attached to.

        :`getter`: Returns the port of this node (read-only).
        """
        return self.__port()

    @property
    def name(self):
        """Name of the node object.

        :`getter`: Returns the name of the node (read-only).
        """
        return self.__name

    @property
    def tag(self):
        """Tagged name of the node object.

        :`getter`: Returns the tagged (user-defined) name of the node (read-only).
        """
        return self.__tag_name

    def _set_tag(self, tag):
        self.__tag_name = tag

    @property
    def _model(self):
        return self.component._model

    @property
    def component(self) -> ModelElement:
        """The component which has ownership of this node.

        :`getter`: Returns the component that this node belongs to (read-only).
        """
        return self.port.component

    def is_neighbour(self, node):
        """Checks if `node` is a connected by an edge to this node.

        Parameters
        ----------
        node : :class:`.Node`
            Node with which to check connection.

        Returns
        -------
        flag : bool
            True if `node` is connected to this node, False otherwise.
        """
        # if not associated with a model yet, check registered connections directly
        try:
            a = self._model.network.has_node(self.full_name)
            b = node.full_name in self._model.network.neighbors(self.full_name)
            return a and b
        except ComponentNotConnected:
            return (
                self.full_name,
                node.full_name,
            ) in self.component._registered_connections.values()

        # if not self.has_model():
        #     return (self, node) in self.__component()._registered_connections.values()
        # return (self._model.network.has_node(self) and
        #         node in self._model.network.neighbors(self))

    @property
    def connections(self):
        """Connections of this node.

        :`getter`: Returns a collection of :class:`.Space`, :class:`.Wire`, or :class:`.DegreeOfFreedom` objects attached to this node (read-only).
        """

        try:
            if self.direction == NodeDirection.INPUT:
                edges = self._model.network.in_edges(self.full_name)
            elif self.direction == NodeDirection.OUTPUT:
                edges = self._model.network.out_edges(self.full_name)
            else:
                edges = self._model.network.edges(self.full_name)

            objects = []
            for edge in edges:
                edge_data = self._model.network.get_edge_data(*edge)
                owner = edge_data["owner"]()
                if isinstance(
                    owner,
                    (
                        components.Space,
                        components.Wire,
                        components.DegreeOfFreedom,
                    ),
                ):
                    objects.append(owner)

            return objects
        except ComponentNotConnected:
            pass

        return tuple()


class SignalNode(Node):
    """Represents a specific small signal degree of freedom. A signal is some small AC
    oscillation in some property, such as longitudinal motion, voltage, laser amplitude,
    etc.

    Parameters
    ----------
    name : str
        Name of the mechanical motion.

    port : :class:`.Port`
        The port that this node belongs to.

    num_frequencies : int
        Number of mechanical frequencies to model
    """

    def __init__(self, name, port, direction, node_type):
        super().__init__(name, port, node_type, direction)
        self.__frequencies = None
        self.__num_frequencies = 1

    @property
    def num_frequencies(self):
        return self.__num_frequencies

    @property
    def frequencies(self):
        if self.__frequencies is None:
            return (self.component._model.fsig.f.ref,)
        else:
            return self.__frequencies

    @frequencies.setter
    def frequencies(self, value):
        self.__frequencies = value
        self.__num_frequencies = len(value)


class OpticalNode(Node):
    """Represents a specific optical port connection at a component.

    OpticalNodes also have additional physical properties such as the beam parameter (of type
    :class:`.BeamParam`) at the nodes' position within the interferometer.

    Parameters
    ----------
    name : str
        Name of the optical node.
    port : :class:`.Port`
        The port that this node belongs to.
    direction : :class:`.NodeDirection`
        True if the field at this node is going into the component.
    """

    def __init__(self, name, port, direction):
        super().__init__(name, port, NodeType.OPTICAL, direction)

        self.__space = None

    def __deepcopy__(self, memo):
        new = super().__deepcopy__(memo)
        # Manually update the weakrefs to be correct
        if self.__space is not None:
            id_space = id(self.__space())
            if id_space in memo:
                new.__space = weakref.ref(memo[id_space])
            else:
                # We need to update this reference later on
                # This will be called when the port property
                # is accessed. When this happens we'll peak back
                # at the memo once it has been filled and get
                # the new port reference. After this the refcount
                # for this function should goto zero and be garbage
                # collected
                def update_later():
                    new.__space = weakref.ref(memo[id_space])

                new.__space = update_later  # just in case something calls
                # this weakref in the meantime
                memo[id(self._model)].after_deepcopy.append(update_later)

        return new

    @staticmethod
    def get_opposite_direction(node):
        """Returns the opposite direction of a node from either a Node object or a full
        string name qualifier for a node, `component.port.direction` `l1.p1.o`.

        Parameters
        ----------
        node : [str | :class:`.Node`]
            Node to invert
        """
        if isinstance(node, str):
            if node.endswith(".o"):
                return node.removesuffix(".o") + ".i"
            elif node.endswith(".i"):
                return node.removesuffix(".i") + ".o"
            else:
                raise ValueError(
                    f"`{node}` string name was not in the form `component.port.i` or `component.port.o`"
                )
        else:
            return node.opposite

    @property
    def opposite(self):
        """The opposite direction node.

        :`getter`: Returns the opposite direction node to this one.
        """
        return getattr(self.port, "o" if self.is_input else "i")

    @property
    def q(self):
        """Beam parameter value at this node.

        :`getter`: Returns the beam parameter at this node. If the beam
                 parameters in the tangential and sagittal planes are
                 different then it returns a tuple of the two parameters.
        :setter: Sets the beam parameter at this node. If the argument
                 provided is a 2-tuple then the parameter is set astigmatically
                 for the node.
        """
        return self.qx, self.qy

    @q.setter
    def q(self, value):
        if is_iterable(value):  # both qx and qy specified
            self.qx = value[0]
            self.qy = value[1]
        else:  # only one q specified
            if self in self._model.gausses:
                self._model.update_gauss(self, qx=value, qy=value)

            else:
                node_name = self.full_name.replace(".", "_")
                name = f"g{node_name}"

                from .gauss import Gauss  # avoids circular import

                self._model.add(Gauss(name, self, qx=value, qy=value))

    @property
    def qx(self):
        """Beam parameter value in the tangential plane.

        :`getter`: Returns the beam parameter at the node in the
                 tangential plane.
        :setter: Sets the beam parameter at the node in the
                 tangential plane.
        """
        trace = self._model.last_trace
        if trace is None:
            gauss = self._model.gausses.get(self)
            # TODO (sjr) Also check in model.cavities for this node as source
            if gauss is None:
                raise RuntimeError(
                    f"No stored beam trace yet performed on model and {self.full_name} "
                    "does not have correspond to any Gauss or Cavity object - unable to "
                    "access any beam parameter at this node."
                )

            warn(
                "No beam trace solution yet stored in the model. Returning "
                "qx based on Gauss object entry."
            )
            qx = gauss.qx
        else:
            qx, _ = trace.get(self, (None, None))
            if qx is None:
                raise RuntimeError(
                    f"Could not find entry for {self.full_name} "
                    "in last stored beam trace solution of the model."
                )

        return qx

    @qx.setter
    def qx(self, value):
        if self in self._model.gausses:
            self._model.update_gauss(self, qx=value)

        else:
            node_name = self.full_name.replace(".", "_")
            name = f"g{node_name}"

            from .gauss import Gauss  # avoids circular import

            self._model.add(Gauss(name, self, q=value))

    @property
    def qy(self):
        """Beam parameter value in the sagittal plane.

        :`getter`: Returns the beam parameter at the node in the
                 sagittal plane.
        :setter: Sets the beam parameter at the node in the
                 sagittal plane.
        """
        trace = self._model.last_trace
        if trace is None:
            gauss = self._model.gausses.get(self)
            # TODO (sjr) Also check in model.cavities for this node as source
            if gauss is None:
                raise RuntimeError(
                    f"No stored beam trace yet performed on model and {self.full_name} "
                    "does not have correspond to any Gauss or Cavity object - unable to "
                    "access any beam parameter at this node."
                )

            warn(
                "No beam trace solution yet stored in the model. Returning "
                "qy based on Gauss object entry."
            )
            qy = gauss.qy
        else:
            _, qy = trace.get(self, (None, None))
            if qy is None:
                raise RuntimeError(
                    f"Could not find entry for {self.full_name} "
                    "in last stored beam trace solution of the model."
                )

        return qy

    @qy.setter
    def qy(self, value):
        if self in self._model.gausses:
            self._model.update_gauss(self, qy=value)

        else:
            node_name = self.full_name.replace(".", "_")
            name = f"g{node_name}"

            from .gauss import Gauss  # avoids circular import

            self._model.add(Gauss(name, self, q=value))

    @property
    def space(self):
        """A reference to the :class:`.Space` object attached to this node.

        :`getter`: Returns a reference to the :class:`.Space`
                   object attached to the node (read-only).
        """
        if self.__space is not None:
            return self.__space()

        try:
            if self.is_input:
                edges = self._model.network.in_edges(self.full_name)
            else:
                edges = self._model.network.out_edges(self.full_name)

            for edge in edges:
                edge_data = self._model.network.get_edge_data(*edge)
                owner = edge_data["owner"]()

                if isinstance(owner, components.Space):
                    self.__space = weakref.ref(owner)
                    return owner
        except ComponentNotConnected:
            pass

        return None



import pytest
from finesse.knm.maps import Map
from finesse.exceptions import FinesseException
import numpy as np


@pytest.fixture
def mock_data():
    x = np.random.rand(3)
    y = np.random.rand(3)
    z = np.random.rand(3, 3)
    return x, y, z


@pytest.mark.parametrize("amp_or_opd", ("amplitude", "opd"))
def test_nans_raise_error(mock_data, amp_or_opd):
    x, y, z = mock_data
    z[1][1] = np.nan
    with pytest.raises(FinesseException):
        Map(x, y, **{amp_or_opd: z})


@pytest.mark.parametrize("amp_or_opd", ("amplitude", "opd"))
def test_wrong_shape(mock_data, amp_or_opd):
    x, y, z = mock_data
    x = np.random.rand(4)
    with pytest.raises(FinesseException):
        Map(x, y, **{amp_or_opd: z})


@pytest.mark.parametrize(
    "auto_kwarg",
    ("auto_remove_tilts", "auto_remove_curvatures", "auto_remove_astigmatism"),
)
def test_auto_args(mock_data, auto_kwarg):
    x, y, z = mock_data
    with pytest.raises(FinesseException):
        Map(x, y, z, **{auto_kwarg: True})



"""Test cases for model component network generation."""


def test_component_network_nodes_represent_model_components(network_model_michelson):
    """Test that the nodes of component networks are the model component names."""
    model, network = network_model_michelson
    assert list(network.nodes) == [component.name for component in model.components]


def test_component_network_edges_link_model_components(network_model_michelson):
    """Test that the edges of component networks link components in the network."""
    model, network = network_model_michelson

    component_names = [component.name for component in model.components]

    for u, v in network.edges:
        assert u in component_names
        assert v in component_names


def test_component_network_edges_contain_connector_data(network_model_michelson):
    """Test that the edges of component networks contain related connector object data."""
    model, network = network_model_michelson

    for _, _, data in network.edges(data=True):
        assert "connection" in data
        # Resolve weakref and check the object is in the model's elements.
        assert data["connection"]() in model.elements.values()


def test_component_network_can_handle_cyclic_models(network_model_sagnac):
    """Test that the component network can handle models with cyclic connections."""
    model, network = network_model_sagnac

    assert len(model.components) == len(network.nodes)



# flake8: noqa
"""Finesse is a Python package for simulating interferometers in the frequency
domain."""


PROGRAM = __name__
DESCRIPTION = "Simulation program for laser interferometers."

# Set the Finesse version.
try:
    from .version import version as __version__
except ImportError:
    raise Exception("Could not find version.py. Ensure you have run setup.")

# Set up some sensible default runtime options.
from .config import configure, autoconfigure

autoconfigure()

# Import a bunch of useful functions and classes into the top-level package.
from .env import (
    is_interactive,
    show_tracebacks,
    tb,
    session_instance as _session_instance,
)
from .constants import values as constants
from .parameter import Parameter, float_parameter
from .gaussian import BeamParam
from .model import Model
from .plotting import init as init_plotting
from .script import syntax, help_ as help
from .script import syntax

# Set up the user session.
session = _session_instance()

from .utilities.storage import save, load
from finesse.utilities.bug_report import bug_report



"""Control system functions."""

import numpy as np
from numpy.polynomial.polynomial import polyvalfromroots


def zpk_fresp(zs, ps, k, w):
    """Evaluate the frequency response of a zpk filter.

    The evaluation is done by multiplying the poles and zeros in pairs which is
    more numerically stable than multiplying all of the zeros and then dividing
    by all of the poles as is done with other tools when there are a large number
    of zeros and poles.

    Parameters
    ----------
    zs : array
        zeros of the filter
    ps : array
        poles of the filter
    k : float
        gain of the filter
    w : array
        frequencies at which to compute the filter
    """
    zs = np.atleast_1d(zs)
    ps = np.atleast_1d(ps)
    s = 1j * w
    tf = k * np.ones_like(s)
    minlen = min(len(zs), len(ps))
    # first multiply the poles and zeros in pairs
    for z, p in zip(zs[:minlen], ps[:minlen]):
        tf *= (s - z) / (s - p)
    # then multiply the remaining poles or zeros
    return tf * polyvalfromroots(s, zs[minlen:]) / polyvalfromroots(s, ps[minlen:])



"""Folder Action."""

from finesse.solutions import BaseSolution
from .base import Action


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Folder(Action):
    """A Folder action collects a new solution every time the action is called.

    An example of this is the 'post step' for the `xaxis`. A folder action is made
    called `post_step` and is passed to a function which will `do` it multiple times.
    After each step the specificed action is called and its solution will be added to
    the folder.
    """

    def __init__(self, name, action, solution):
        super().__init__(name)
        self.action = action
        self.folder_solution = BaseSolution(name)
        solution.add(self.folder_solution)

    def _do(self, state):
        sol = state.apply(self.action)
        if sol:
            self.folder_solution.add(sol)

    def _requests(self, model, memo, first=True):
        return self.action._requests(model, memo)



"""Material objects are a simple structure that contain commonly used
properties for calculations. By default this includes Fused Silica and
Silicon at 123K.

.. todo::

    At some point add in wavelenth and temperature dependence to these
    Material objects.
"""
cimport cython

@cython.auto_pickle(True)
cdef class Material(object):
    """"""
    def __init__(self, alpha, nr, dndT, kappa, emiss, poisson, E, rho, C, T):
        self.values.alpha   = alpha
        self.values.nr      = nr
        self.values.dndT    = dndT
        self.values.kappa   = kappa
        self.values.emiss   = emiss
        self.values.poisson = poisson
        self.values.E       = E
        self.values.T       = T
        self.values.C       = C
        self.values.rho     = rho

    @property
    def alpha(self):
        r"""Thermal expansion coefficient :math:`\alpha`"""
        return self.values.alpha

    @property
    def nr(self):
        r"""Refractive index :math:`n`"""
        return self.values.nr

    @property
    def dndT(self):
        r"""Thermo refractive coefficient :math:`\frac{dn}{dT} \: [K^{-1}]`"""
        return self.values.dndT

    @property
    def kappa(self):
        r"""Thermal conductivity :math:`\kappa \: [W m^{-1}]`"""
        return self.values.kappa

    @property
    def emiss(self):
        r"""Emissivity"""
        return self.values.emiss

    @property
    def poisson(self):
        r"""Poisson's ratio"""
        return self.values.poisson

    @property
    def E(self):
        r"""Young's modulus :math:`[\mathrm{kg} \, m^{-3}]`"""
        return self.values.E

    @property
    def rho(self):
        r"""Density :math:`[\mathrm{kg} \, m^{-3}]`"""
        return self.values.rho

    @property
    def C(self):
        r"""specific heat :math:`[J \, \mathrm{kg}^{-1}]`"""
        return self.values.C

    @property
    def T(self):
        r"""The temperature material's properties are defined at :math:`[K]`"""
        return self.values.T

Vacuum = Material(
    0, # thermal expansion coefficient
    1, # refractive index
    0, # dn/dt
    0, # thermal conductivity
    0, # emissivity
    0, # Poisson's ration
    0, # Young's modulus
    0, # density
    0, # specific heat
    0, # reference temperature
)

FusedSilica = Material(
    5.5e-7, # thermal expansion coefficient
    1.45,   # refractive index
    8.6e-6, # dn/dt
    1.38,   # thermal conductivity
    0.91,   # emissivity
    0.17,   # Poisson's ration
    7.2e10, # Young's modulus
    2202,   # density
    772,    # specific heat
    297,    # reference temperature
)

Gold = Material(
    14.13e-6, # thermal expansion coefficient
    0.258,    # refractive index
    0,        # dn/dt
    315,      # thermal conductivity
    0.05,     # emissivity
    0.42,     # Poisson's ration
    7.9e10,   # Young's modulus
    1930,     # density
    129,      # specific heat
    297,      # reference temperature
)
# Taken from Voyager GWINC
Silicon123K_sum = Material(
    1e-10,   # thermal expansion coefficient
    3.4,     # refractive index @ 2um
    1e-4,    # dn/dt
    700,     # thermal conductivity
    0.7,     # emissivity, https://www.sciencedirect.com/science/article/pii/S0017931019361289
    0.27,    # Poisson's ration
    155.8e9, # Young's modulus
    2329,    # density
    300,     # specific heat
    123,     # reference temperature
)

CaF2_300K_2um = Material(
    18.5e-6, # thermal expansion coefficient
    1.4239,  # refractive index @ 2um
    -10e-6,  # dn/dt
    9.71,    # thermal conductivity
    0.88,    # emissivity
    0.26,    # Poisson's ration
    75.8e9,  # Young's modulus
    3180,    # density
    854,     # specific heat
    300,     # reference temperature
)

# Fused silica used for iLIGO
# reference CTE (NIST SRM  739)
Corning7940 = Material(
    4.869e-07, # thermal expansion coefficient
    1.45,      # refractive index
    9.6e-6,    # dn/dt
    1.367,     # thermal conductivity
    0.9,       # emissivity
    0.167,     # Poisson's ration
    72.93E9,   # Young's modulus
    2220.00,   # density
    704.21,    # specific heat
    293.15,    # reference temperature
)

# Suprasil ref
# n, Thermal expansion coefficient from data sheet
# https://www.heraeus.com/media/media/hca/doc_hca/products_and_solutions_8/optics/Data_and_Properties_Optics_fused_silica_EN.pdf
suprasil3002_2um = Material(
    5.9e-7,  # thermal expansion coefficient
    1.499,   # refractive index
    8.89e-6, # dn/dt
    1.38,    # thermal conductivity
    0.9,     # emissivity
    0.17,    # Poisson's ration
    73.1E9,  # Young's modulus
    2203,    # density
    964,     # specific heat
    293.15,  # reference temperature
)

BK7_2um = Material(
    7.1e-6,  # thermal expansion coefficient
    1.4946,  # refractive index
    8.94E-7, # dn/dt
    1.114,   # thermal conductivity
    0.9,     # emissivity
    0.206,   # Poisson's ration
    82e9,    # Young's modulus
    2510,    # density
    858,     # specific heat
    293.15,  # reference temperature
)

#   zblan reference   Zhu X., Peyghambarian N., High-Power ZBLAN Glass Fiber Lasers: Review and Prospect, 2010
ZBLAN_2um = Material(
    17.2e-6,   # thermal expansion coefficient
    1.4956,    # refractive index
    -14.75e-6, # dn/dt
    0.628,     # thermal conductivity
    0.9,       # emissivity
    0.206,     # Poisson's ration
    58.5e9,    # Young's modulus
    4330,      # density
    151,       # specific heat
    293.15,    # reference temperature
)



""""""
import numpy as np
from finesse.cymath.sparsemath import CSRMatrix
from scipy.sparse import random


def test_sparse_csr_multiply():
    for _ in range(1000):
        N = np.random.randint(1, 10)
        M = np.random.randint(1, 10)
        A = random(M, N, dtype=complex, density=0.25).todense()
        x = np.random.randint(-100, 100, size=N)
        csr = CSRMatrix(A)
        # These should be numerically exact... but apprently not on debian test server
        assert np.max(abs(csr.multiply(x) - (A @ np.atleast_2d(x).T).T)) < 1e-6


def test_sparse_csr_vector_matrix_vector():
    for _ in range(1000):
        N = np.random.randint(1, 10)
        M = np.random.randint(1, 10)
        A = random(M, N, dtype=complex, density=0.5).todense()
        x = np.random.randint(-100, 100, size=N) + 1j * np.random.randint(
            -100, 100, size=N
        )
        y = np.random.randint(-100, 100, size=M) + 1j * np.random.randint(
            -100, 100, size=M
        )
        csr = CSRMatrix(A)
        # These should be numerically exact... but apprently not on debian test server
        assert (
            np.max(
                abs(csr.zcsrgevmv(x, y) - np.atleast_2d(y) @ (A @ np.atleast_2d(x).T))
            )
            < 1e-6
        )


def test_sparse_csr_vector_hermitian_matrix_vector():
    for _ in range(1000):
        N = np.random.randint(1, 10)
        M = np.random.randint(1, 10)
        A = random(M, N, dtype=complex, density=0.5).todense()
        x = np.random.randint(-100, 100, size=N) + 1j * np.random.randint(
            -100, 100, size=N
        )
        y = np.random.randint(-100, 100, size=M) + 1j * np.random.randint(
            -100, 100, size=M
        )
        csr = CSRMatrix(A)
        # These should be numerically exact... but apprently not on debian test server
        assert (
            np.max(
                abs(
                    csr.zcsrgecmv(x, y)
                    - np.atleast_2d(y).conjugate() @ (A @ np.atleast_2d(x).T)
                )
            )
            < 1e-6
        )



"""Test fixtures shared by all tests."""

import locale
import os
import sys
from pathlib import Path

import pytest
from faker import Faker

from finesse.env import warn

# Add the test root directory to the path so test scripts can import the testutil
# package directly.
sys.path.append(str(Path(__file__).parent))
# # add the repo root directory to path so tests can import the `scripts` folder
sys.path.append(str(Path(__file__).parent.parent))


# Register test utilities package assertion error rewriting.
pytest.register_assert_rewrite(
    "testutils.cli", "testutils.diff", "testutils.fuzzing", "testutils.text"
)


# Set up a repeatable Faker seed.
FAKER_SEED = 314159
Faker.seed(FAKER_SEED)

# list of platforms for custom platform marker
PLATFORMS = set(["aix", "emscripten", "linux", "wasi", "win32", "cygwin", "darwin"])


def pytest_addoption(parser):
    import hypothesis

    # Add an option to change the Hypothesis max_examples setting via the `pytest` CLI.
    parser.addoption(
        "--hypothesis-max-examples",
        action="store",
        default=None,
        help="set the Hypothesis max_examples setting",
    )
    # Add an option to change the Hypothesis suppress_health_check setting via the `pytest` CLI.
    parser.addoption(
        "--hypothesis-suppress-health-check",
        action="store",
        # Work around failing fuzzing healthchecks:
        # "It looks like your strategy is filtering out a lot of data."
        # TODO: figure out why it is producing too much and then reset to None here.
        default=(hypothesis.HealthCheck.filter_too_much,),
        # default=None,
        help="set the Hypothesis suppress_health_check setting as list of int",
    )


def pytest_configure(config):
    import hypothesis

    # Set Hypothesis max_examples and suppress_health_check  using what, if
    # anything, was specified in the call to the `pytest` CLI.
    hypothesis_max_examples = config.getoption("--hypothesis-max-examples")
    hypothesis_suppress_health_check = config.getoption(
        "--hypothesis-suppress-health-check"
    )

    if hypothesis_max_examples is not None:
        max_examples = int(hypothesis_max_examples)
    else:
        max_examples = hypothesis.settings.default.max_examples

    if hypothesis_suppress_health_check is not None:
        # cmdline arg is string, default could be directly list of HealthCheck
        if isinstance(hypothesis_suppress_health_check, str):
            suppress_health_check = (
                hypothesis.HealthCheck(int(i))
                for i in hypothesis_suppress_health_check.split(",")
            )
        else:
            suppress_health_check = hypothesis_suppress_health_check
    else:
        suppress_health_check = hypothesis.settings.default.suppress_health_check

    hypothesis.settings.register_profile(
        "hypothesis-overridden",
        max_examples=max_examples,
        suppress_health_check=suppress_health_check,
        # Do not enable deadline setting for windows CI because it results in random test failures
        deadline=(
            None if os.getenv("windows_ci") else hypothesis.settings.default.deadline
        ),
    )
    hypothesis.settings.load_profile("hypothesis-overridden")

    # store the numeric locale for reversion if a test unexpectedly changes it
    pytest.expected_locale = locale.getlocale(locale.LC_NUMERIC)
    for plat in PLATFORMS:
        config.addinivalue_line("markers", f"{plat}: mark test to run only on {plat}")


def pytest_xdist_auto_num_workers(config):
    import psutil

    # There are diminishing returns after using more than 3 cores
    if psutil.cpu_count(logical=False) > 2:
        return 3
    elif psutil.cpu_count(logical=False) > 1:
        return 2
    else:
        return 0


def pytest_runtest_teardown(item, nextitem):
    # check if the numeric locale has changed
    curr_locale = locale.getlocale(locale.LC_NUMERIC)

    if curr_locale != pytest.expected_locale:
        warn(
            f"locale unexpectedly changed from {pytest.expected_locale} to {curr_locale}",
            RuntimeWarning,
        )

        # numeric locale has changed, revert it back
        locale.setlocale(locale.LC_NUMERIC, pytest.expected_locale)


@pytest.fixture(autouse=True)
def test_setup_and_teardown():
    # Set up global state before, and clean up after, each test.
    from finesse.config import autoconfigure
    from finesse.datastore import invalidate

    # Reset to default configuration.
    autoconfigure()

    # Run the test.
    yield

    # Delete the global cache.
    invalidate()


# https://docs.pytest.org/en/7.1.x/example/markers.html#marking-platform-specific-tests-with-pytest
def pytest_runtest_setup(item):
    plat = sys.platform
    test_platform_markers = PLATFORMS.intersection(
        mark.name for mark in item.iter_markers()
    )
    if test_platform_markers and plat not in test_platform_markers:
        pytest.skip(f"Can not run test on platform {plat}")



"""Collection of Actions that involve beam tracing."""

from .base import Action

import logging

LOGGER = logging.getLogger(__name__)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class ABCD(Action):
    """Computation of an ABCD matrix over a path.

    See :func:`.compute_abcd` for details.
    """

    def __init__(self, name="abcd", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.ABCD(solution_name=self.name, **self.kwargs)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class BeamTrace(Action):
    """Full beam tracing on a complete model.

    See :meth:`.Model.beam_trace` for details.
    """

    def __init__(self, name="beam_trace", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.beam_trace(solution_name=self.name, **self.kwargs)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class PropagateBeam(Action):
    """Propagation of a beam, in a single plane, through a given path.

    See :meth:`.Model.propagate_beam` for details.
    """

    def __init__(self, name="propagation", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.propagate_beam(solution_name=self.name, **self.kwargs)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class PropagateAstigmaticBeam(Action):
    """Propagation of a beam, in both planes, through a given path.

    See :meth:`.Model.propagate_beam_astig` for details.
    """

    def __init__(self, name="astig_propagation", **kwargs):
        super().__init__(name)
        self.kwargs = kwargs

    def _requests(self, model, memo, first=True):
        pass

    def _do(self, state):
        return state.model.propagate_beam_astig(solution_name=self.name, **self.kwargs)



cimport numpy as np
from finesse.cymath.complex cimport complex_t, DenseZMatrix


cdef class KnmMatrix:
    cdef readonly:
        np.ndarray data
        np.ndarray modes
        np.ndarray homs
        str name

    cdef:
        complex_t[:, ::1] data_view
        DenseZMatrix mtx
        const int[:, ::1] modes_view

    cdef (Py_ssize_t, Py_ssize_t) field_indices_from(self, key)
    cdef complex_t coupling(self, int n1, int m1, int n2, int m2) noexcept nogil


cpdef make_unscaled_X_scatter_knm_matrix(int[:,::1] modes)
cpdef make_unscaled_Y_scatter_knm_matrix(int[:,::1] modes)


# Compute loss from scattering for each coupling,
# required for  quantum noise calculations
cdef void knm_loss(const complex_t* knm_mat, double* out, Py_ssize_t N) noexcept nogil

cdef void c_zero_tem00_phase(
    const complex_t[:, ::1] knm_mat,
    complex_t[:, ::1] out
) noexcept nogil

cdef void c_flip_odd_horizontal(
    DenseZMatrix *knm_mat,
    const int[:, ::1] homs
) noexcept nogil

cdef void c_reverse_gouy_phases(
    double x_gouy1, double y_gouy1,
    double x_gouy2, double y_gouy2,
    const complex_t[:, ::1] knm_mat,
    const int[:, ::1] homs,
    complex_t[:, ::1] out
) noexcept nogil

cdef complex_t rev_gouy(
    double x_gouy1,
    double y_gouy1,
    double x_gouy2,
    double y_gouy2,
    complex_t k,
    int n1, int m1, int n2, int m2,
) noexcept nogil



from finesse.cymath.math cimport ceil, floor
from finesse.cymath.math cimport nmin

IF FINESSE3_DISABLE_OPENMP == 0:
    cimport openmp

cpdef int determine_nthreads_even(int pts, int divisor) noexcept nogil:
    """Determine the number of threads to use for a routine
    with an outer loop of size `pts`.

    Parameters
    ----------
    pts : int
        Size of outer loop of routine to be parallelised.

    divisor : int
        Estimate of scaling for number of threads where the
        nominal thread count will then be ``pts // divisor``.

    Returns
    -------
    nthreads : int
        Determined thread count. Or 1 if no openmp support
        available.

    Notes
    -----
    The `divisor` arg defines the nominal number of threads
    via ``nominal = pts // divisor``. Hence, this argument
    should correspond to the best-estimate of multi-threaded
    performance scaling behaviour of the routine to follow.

    The actual value returned will be either unity or an even
    number (depending on values of `pts` and `divisor`) which
    is clipped by the maximum number of threads as given by
    the OpenMP routine ``omp_get_max_threads``.

    If Finesse was compiled without OpenMP support then this
    always returns 1.
    """
    IF FINESSE3_DISABLE_OPENMP == 1:
        return 1
    ELSE:
        # Nominal number of threads as pts // divisor
        cdef int nominal = int(ceil(pts / float(divisor)))
        # If nominal is not unity then ensure it's even
        if nominal != 1:
            nominal = int(2 * floor(nominal / 2.0))

        # Note that if OMP_NUM_THREADS is set then omp_get_max_threads
        # will return the value defined in that variable
        return nmin(nominal, openmp.omp_get_max_threads())



from hypothesis import given, assume, settings, HealthCheck
from hypothesis.strategies import floats, complex_numbers
from testutils.fuzzing import DEADLINE, c_integer


@given(value=c_integer)
@settings(
    deadline=DEADLINE,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
)
def test_integer_fuzzing(fuzz_argument, value):
    assert fuzz_argument(value) == value


@given(value=floats(allow_nan=False, allow_infinity=True))
@settings(
    deadline=DEADLINE,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
)
def test_float_fuzzing(fuzz_argument, value):
    assume(not str(value).endswith("e"))
    assert fuzz_argument(value) == value


@given(value=complex_numbers(allow_nan=False, allow_infinity=False))
@settings(
    deadline=DEADLINE,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
)
def test_complex_fuzzing(fuzz_argument, value):
    assert fuzz_argument(value) == value



"""Temporary actions allow for temporary state changes to perform some chosen set of
actions with, then returning to the original state."""

import logging

from ...parameter import GeometricParameter
from .base import Action, convert_str_to_parameter

LOGGER = logging.getLogger(__name__)


def temporary(action):
    """Converts an action into a temporary action.

    This function takes a target action, and returns an action that
    takes multiple actions as arguments. When the returned action is
    run, it will first run the target action, then all actions passed to
    it, then restore the changes made by the target action, e.g.

    .. code-block:: python

        temporary(Change({'m1.phi': 10}))(
            Xaxis(l1.P, 'lin', 0, 10, 100)
        )

    will first set the parameter `m1.phi` to 10, then run a sweep of
    `l1.P`, then restore `m1.phi` to its previous value.

    Parameters
    ----------
    action : Action
        The action to make temporary.

    Returns
    -------
    action
        An action that temporarily applies the passed action when run.
    """

    def func(*args, **kwargs):
        return Temporary(action, *args, **kwargs)

    return func


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class Temporary(Action):
    """Make the first action in a series of actions temporary, i.e. restore its
    parameters after the rest of the actions are complete."""

    def __init__(self, temp_action, *actions):
        super().__init__("series", True)
        self.temp_action = temp_action
        self.actions = actions

    def _do(self, state):
        rq = self.temp_action.get_requests(state.model)
        params = {
            convert_str_to_parameter(state.model, p): convert_str_to_parameter(
                state.model, p
            ).value
            for p in rq["changing_parameters"]
        }

        state.apply(self.temp_action)

        curr_sol = None
        for action in self.actions:
            next_sol = state.apply(action)
            if next_sol and not curr_sol:
                first = next_sol  # need to return the first one
            if next_sol:
                if curr_sol:
                    curr_sol.add(next_sol)
                curr_sol = next_sol

        for param, value in params.items():
            param.value = value
            param._reset_cvalue()

        return first

    def _requests(self, model, memo, first=True):
        self.temp_action._requests(model, memo, False)
        for action in self.actions:
            action._requests(model, memo, False)


# IMPORTANT: renaming this class impacts the katscript spec and should be avoided!
class TemporaryParameters(Action):
    """An action that will revert any changed parameters back to their values before
    this action was called. Options exist to include or exclude certain Parameters from
    this reversion. This action does not generate any Solution.

    Parameters
    ----------
    action : :class:`finesse.analysis.actions.base.Action`
        Action to perform followed by reverting requested Parameters in the model

    include : [iterable|str], optional
        Parameters that *should* be included.

        If a single string is given it can be a Unix file style wildcard (See ``fnmatch``).
        A value of None means everything is included.

        If an iterable is provided it must be a list of names or Parameter objects.

    exclude : [iterable|str], optional
        Parameters that *should not* be included.

        If a single string is given it can be a Unix file style wildcard (See ``fnmatch``).
        A value of None means nothing is excluded.

        If an iterable is provided it must be a list of names or Parameter objects.
    """

    def __init__(self, action, *, include=None, exclude=None):
        if action is None:
            raise ValueError("Action must be provided")

        super().__init__("temp_param")
        self.action = action
        self.include = include
        self.exclude = exclude

    def _do(self, state):
        params = state.sim.model.get_parameters(
            include=self.include,
            exclude=self.exclude,
            are_changing=True,
            are_symbolic=False,
        )
        initial = {p: p.value for p in params}
        # apply the action
        sol = state.apply(self.action)

        # Start resetting parameter values and updating things
        for param, prev in initial.items():
            param.value = prev
        # Ensure the __cvalue of each symbolic parameter gets reset accordingly
        for param in state.sim.changing_parameters:
            param._reset_cvalue()

        if any(
            type(p) is GeometricParameter and p.is_symbolic
            for p in state.sim.changing_parameters
        ):
            state.model._update_symbolic_abcds()

        return sol

    def _requests(self, model, memo, first=True):
        self.action._requests(model, memo)
